from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Optional

import numpy as np
import toloka.client as toloka

from .. import evaluation

from .pipeline import FeedbackLoop
from .results import get_results


@dataclass
class Metrics:
    @dataclass
    class QualityDistribution:
        unknown: int
        ok: int
        not_ok: int

    markup_attempts: Dict[int, int]
    confidence_percentiles: Dict[int, float]
    quality: QualityDistribution

    cost_markup: float
    cost_check: float
    duration_seconds: float

    cost_markup_per_unit: Optional[float]
    cost_check_per_unit: Optional[float]
    session_speed: Optional[float]

    # TODO: добавить objects_speed, аналогичную нашей records_speed
    # TODO: стоимость не учитывает неполные страницы заданий


percentiles = [10, 25, 50, 75, 90, 95, 99]


def calculate_metrics(
    fb_loop: FeedbackLoop,
    markup_pool_id: str,
    check_pool_id: str,
) -> Metrics:
    markup_pool = fb_loop.client.get_pool(markup_pool_id)
    check_pool = fb_loop.client.get_pool(check_pool_id)

    assert all(
        pool.status == toloka.Pool.Status.CLOSED for pool in (markup_pool, check_pool)
    ), 'pools must be closed before metrics calculation'

    duration_seconds = (check_pool.last_stopped - markup_pool.created).total_seconds()

    markup_assignments = fb_loop.get_human_markups(markup_pool_id)
    accepted_markup_assignments = [
        assignment for assignment, _ in markup_assignments if assignment.status == toloka.Assignment.ACCEPTED
    ]
    accepted_check_assignments = list(
        fb_loop.client.get_assignments(status=toloka.Assignment.ACCEPTED, pool_id=check_pool_id)
    )

    solution_id_to_evaluation = evaluation.collect_evaluations_from_check_assignments(
        assignments=accepted_check_assignments,
        check_task_mapping=fb_loop.check_task_mapping,
        pool_input_objects=None,
        aggregation_algorithm=fb_loop.params.evaluation.aggregation_algorithm,
        confidence_threshold=fb_loop.params.quality.confidence_threshold,
    )

    given_bonuses = evaluation.calculate_bonuses_for_pool_of_markup_assignments(
        markup_assignments=markup_assignments,
        solution_id_to_evaluation=solution_id_to_evaluation,
        check_task_mapping=fb_loop.check_task_mapping,
        control_params=fb_loop.params.control_markup,
    )

    cost_markup_bonus = sum(bonus.amount for bonus in given_bonuses)
    cost_markup = len(accepted_markup_assignments) * markup_pool.reward_per_assignment + cost_markup_bonus
    cost_check = len(accepted_check_assignments) * check_pool.reward_per_assignment

    cost_markup_per_unit = None
    cost_check_per_unit = None
    session_speed = None

    # TODO: bring volume back, broken after ASREXP-1687
    # if fb_loop.objects[0].volume() is not None:
    #     total_volume = sum(obj.volume() for obj in fb_loop.objects)
    #     cost_markup_per_unit = cost_markup / total_volume
    #     cost_check_per_unit = cost_check / total_volume
    #     session_speed = total_volume / duration_seconds

    markup_attempts = defaultdict(int)
    for attempts in evaluation.get_tasks_attempts(markup_assignments, fb_loop.markup_task_mapping).values():
        markup_attempts[attempts] += 1

    confidences = []
    quality_unknown = 0
    quality_ok = 0
    quality_not_ok = 0
    for task_result in get_results(
        fb_loop.pool_input_objects,
        markup_assignments,
        solution_id_to_evaluation,
        fb_loop.markup_task_mapping,
        fb_loop.check_task_mapping,
    ):
        task_confidences = [
            None if solution.evaluation is None else solution.evaluation.confidence
            for solution in task_result.solutions
        ]
        confidences += [c for c in task_confidences if c is not None]

        max_confidence = None
        for c in task_confidences:
            if c is None:
                continue
            if max_confidence is None:
                max_confidence = c
            elif c > max_confidence:
                max_confidence = c

        if max_confidence is None:
            quality_unknown += 1
        else:
            if max_confidence >= fb_loop.params.quality.confidence_threshold:
                quality_ok += 1
            else:
                quality_not_ok += 1

    return Metrics(
        markup_attempts=markup_attempts,
        confidence_percentiles={p: np.percentile(np.array(confidences), p) for p in percentiles},
        quality=Metrics.QualityDistribution(unknown=quality_unknown, ok=quality_ok, not_ok=quality_not_ok),
        cost_markup=cost_markup,
        cost_check=cost_check,
        duration_seconds=duration_seconds,
        cost_markup_per_unit=cost_markup_per_unit,
        cost_check_per_unit=cost_check_per_unit,
        session_speed=session_speed,
    )



from collections import defaultdict
from dataclasses import dataclass
import enum
from typing import List, Dict, Optional

import toloka.client as toloka

from .. import evaluation, mapping, worker as workers


class SolutionVerdict(enum.Enum):
    BAD = 0
    UNKNOWN = 1
    OK = 2


@dataclass
class Solution:
    solution: mapping.Objects
    verdict: SolutionVerdict
    evaluation: Optional[evaluation.SolutionEvaluation]
    assignment_accuracy: float  # ratio of correctly completed tasks in source assignment
    assignment_evaluation_recall: float  # ratio of evaluated tasks in source assignment
    worker: workers.Worker

    def __lt__(self, other: 'Solution') -> bool:
        if self.verdict.value != other.verdict.value:
            return self.verdict.value < other.verdict.value

        assert (
            self.evaluation is None
            and other.evaluation is None
            or self.evaluation is not None
            and other.evaluation is not None
        )

        if self.evaluation is not None and self.evaluation.confidence != other.evaluation.confidence:
            return self.evaluation.confidence < other.evaluation.confidence

        def f1(p, r):
            return 2 * (p * r) / (p + r)

        return f1(self.assignment_accuracy, self.assignment_evaluation_recall) < f1(
            other.assignment_accuracy, other.assignment_evaluation_recall
        )


def get_results(
    pool_input_objects: List[mapping.Objects],
    markup_assignments: List[mapping.AssignmentSolutions],
    solution_id_to_evaluation: Dict[mapping.TaskID, evaluation.SolutionEvaluation],
    markup_task_mapping: mapping.TaskMapping,
    check_task_mapping: mapping.TaskMapping,
) -> List[List[Solution]]:
    assert all(
        assignment.status in (toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED)
        for assignment, _ in markup_assignments
    )

    assignment_accuracy_evaluation_strategy = evaluation.CheckAssignmentAccuracyEvaluationStrategy(
        solution_id_to_evaluation, check_task_mapping
    )
    assignments_evaluations = [
        assignment_accuracy_evaluation_strategy.evaluate_assignment(assignment) for assignment in markup_assignments
    ]
    assignment_id_to_accuracy = {e.assignment.id: e.get_accuracy() for e in assignments_evaluations}
    assignment_id_to_evaluation_recall = {e.assignment.id: e.get_evaluation_recall() for e in assignments_evaluations}

    task_id_to_result = defaultdict(list)
    for assignment, solutions in markup_assignments:
        for input_objects, output_objects in solutions:
            task_id = markup_task_mapping.task_id(input_objects)
            solution_id = check_task_mapping.task_id(input_objects + output_objects)
            solution_evaluation = solution_id_to_evaluation.get(solution_id)
            verdict = SolutionVerdict.UNKNOWN
            if solution_evaluation:
                verdict = SolutionVerdict.OK if solution_evaluation.ok else SolutionVerdict.BAD
            solution = Solution(
                solution=output_objects,
                worker=workers.Human(assignment),
                verdict=verdict,
                evaluation=solution_evaluation,
                assignment_accuracy=assignment_id_to_accuracy[assignment.id],
                assignment_evaluation_recall=assignment_id_to_evaluation_recall[assignment.id],
            )
            task_id_to_result[task_id].append(solution)

    for solutions in task_id_to_result.values():
        solutions.sort(reverse=True)

    # order results in pool input order
    results = []
    for input_objects in pool_input_objects:
        results.append(task_id_to_result[markup_task_mapping.task_id(input_objects)])

    return results



from functools import lru_cache
from typing import List, Optional


from .. import classification, classification_loop

from .base import OptionType


def get_rec_label(obj: OptionType, recommended_range: List[OptionType]) -> str:
    if isinstance(obj, str):
        return f'Recommended range: {recommended_range}'
    elif isinstance(obj, float):
        return f'Recommended range: [{recommended_range[0]:.2f}, {recommended_range[-1]:.2f}]'
    elif isinstance(obj, bool) and len(recommended_range) == 1:
        return f'Recommended: {recommended_range[0]}'
    else:
        return f'Recommended range: [{recommended_range[0]}, {recommended_range[-1]}]'


@lru_cache(maxsize=None)
def get_overlap(overlap: str, confidence: Optional[float] = None) -> classification_loop.Overlap:
    spl = overlap.split('-')
    if len(spl) == 1:
        ovr = int(overlap)
        assert ovr >= 1
        return classification_loop.StaticOverlap(ovr)
    min_overlap, max_overlap = map(int, spl)
    assert 0 < min_overlap <= max_overlap <= 100
    if min_overlap == max_overlap:
        return classification_loop.StaticOverlap(min_overlap)
    return classification_loop.DynamicOverlap(min_overlap, max_overlap, confidence)


def overlap_is_static(overlap: str) -> bool:
    return isinstance(get_overlap(overlap, 0.5), classification_loop.StaticOverlap)


def get_agg_algorithm(agg: str) -> classification.AggregationAlgorithm:
    return {
        'DS': classification.AggregationAlgorithm.DAWID_SKENE,
        'MV': classification.AggregationAlgorithm.MAJORITY_VOTE,
        'ML': classification.AggregationAlgorithm.MAX_LIKELIHOOD,
    }[agg]



from dataclasses import dataclass

from .common import LocalizedString, DEFAULT_LANG
from .functions import TaskFunction


@dataclass
class TaskSpec:
    id: str
    function: TaskFunction
    name: LocalizedString
    description: LocalizedString
    instruction: LocalizedString

    def __post_init__(self):
        LocalizedString.convert_dataclass(self)
        assert DEFAULT_LANG in self.name.lang_to_text, f'please provide name for {DEFAULT_LANG} language'



from dataclasses import dataclass
from typing import Dict, Optional

from lzy.api import op, LzyRemoteEnv
from lzy.api.whiteboard import whiteboard, view
import toloka.client as toloka

from . import registration

# TODO: E2E test: register -> revoke -> ... -> get [for different (task_id, lang) cases]


@whiteboard(tags=['expert_registration'], namespace='default')
@dataclass
class Whiteboard:
    pool: toloka.Pool
    task_id: str
    lang: str
    name_to_code: Dict[str, str]
    name_to_worker_id: Dict[str, str] = None


@op
def approve_expected_codes(client: toloka.TolokaClient, pool_id: str, name_to_code: Dict[str, str]) -> Dict[str, str]:
    return registration.approve_expected_codes(client, pool_id, name_to_code)


def register_experts(
    client: toloka.TolokaClient,
    name_to_code: Dict[str, str],
    pool: toloka.Pool,
    task_id: Optional[str] = None,
    lang: Optional[str] = None,
) -> Dict[str, str]:
    wb = Whiteboard(pool=pool, task_id=task_id, lang=lang, name_to_code=name_to_code)

    env = LzyRemoteEnv()
    with env.workflow('expert_registration', whiteboard=wb):
        wb.name_to_worker_id = approve_expected_codes(client, pool.id, name_to_code)

    return wb.name_to_worker_id


def get_registered_experts(
    client: toloka.TolokaClient,
    task_id: Optional[str] = None,
    lang: Optional[str] = None,
) -> Dict[str, str]:
    # in whiteboards, we store information about expert names, but we also could revoke some registrations, so we have
    # to match registrations from whiteboards with user skills
    env = LzyRemoteEnv()
    name_to_worker_id = {}
    for whiteboard in env.whiteboards([Whiteboard]):
        if task_id and whiteboard.task_id != task_id:
            continue
        if lang and whiteboard.lang != lang:
            continue
        name_to_worker_id.update(whiteboard.name_to_worker_id)
    skills = registration.get_skills(task_id, lang, client)
    worker_ids = {}
    for skill in skills:
        worker_ids |= [user_skill.user_id for user_skill in client.get_user_skills(skill_id=skill.id)]
    return {name: worker_id for name, worker_id in name_to_worker_id.items() if worker_id in worker_ids}



from dataclasses import dataclass, field
import difflib
import json
import logging
from typing import Tuple, Optional

from bs4 import BeautifulSoup
import toloka.client as toloka

from .. import base, experts, mapping, instruction, project

from .utils import ask
from ..base import TaskSpec

logger = logging.getLogger(__name__)


@dataclass
class PreparedTaskSpec:
    task_spec: TaskSpec
    lang: str
    scenario: project.Scenario = project.Scenario.DEFAULT

    view: toloka.project.view_spec.ViewSpec = field(init=False)
    task_mapping: mapping.TaskMapping = field(init=False)
    dumped_view: str = field(init=False)

    def __post_init__(self):
        if isinstance(self, AnnotationTaskSpec):
            builder = project.AnnotationBuilder
        else:
            builder = {
                base.ClassificationFunction: project.ClassificationBuilder,
                base.SbSFunction: project.SbSBuilder,
                base.AnnotationFunction: project.AnnotationCheckBuilder,
            }[type(self.function)]

        self.view, self.task_mapping = builder(task_function=self.function, lang=self.lang).get_scenario(
            self.scenario, self.for_preview()
        )
        self.dump_view()

    def dump_view(self):
        self.dumped_view = json.dumps(json.loads(self.view.unstructure()['config']), ensure_ascii=False)

    @property
    def id(self) -> str:
        return self.task_spec.id

    @property
    def function(self) -> base.TaskFunction:
        return self.task_spec.function

    @property
    def name(self) -> base.LocalizedString:
        if self.scenario == project.Scenario.DEFAULT:
            return self.task_spec.name
        return self.task_spec.name + ' (' + self.scenario.project_name_suffix + ')'

    @property
    def description(self) -> base.LocalizedString:
        return self.task_spec.description

    @property
    def instruction(self) -> base.LocalizedString:
        return self.task_spec.instruction

    def log_message(self) -> str:
        scenario = '' if self.scenario == project.Scenario.DEFAULT else f' to {self.scenario.name}'
        return f'task "{self.task_spec.id}" in {self.lang}{scenario}'

    def for_preview(self) -> bool:
        return False

    def overload_view(self, view: toloka.project.view_spec.ViewSpec):
        self.view = view
        self.dump_view()


@dataclass
class PreviewTaskSpec(PreparedTaskSpec):
    def for_preview(self) -> bool:
        return True


@dataclass
class AnnotationTaskSpec(PreparedTaskSpec):
    check: PreparedTaskSpec = field(init=False)

    def __post_init__(self):
        super().__post_init__()
        check_spec = TaskSpec(
            id=f'{self.task_spec.id}_check',
            function=self.task_spec.function,
            name=self.task_spec.name + ' – ' + project.CHECK,
            description=self.description + ' – ' + project.CHECK,
            instruction=self.instruction,
        )
        self.check = PreparedTaskSpec(check_spec, self.lang, self.scenario)


@dataclass
class AnnotationPreviewTaskSpec(PreviewTaskSpec, AnnotationTaskSpec):
    def __post_init__(self):
        super().__post_init__()
        # we don't use it for previews directly, but we need transformed task mapping in check spec here
        self.check = PreviewTaskSpec(self.check.task_spec, self.lang, self.scenario)


# Toloka localized projects are not suitable for us, because:
# - in task view, only labels can be localized in normal way without if/else conditions, i.e. we can't change output
#   validation depends on language for example, see
#   https://toloka.ai/ru/docs/guide/concepts/project-languages.html#project-languages__interface-translate.
# - Worker can choose multiple languages, on which he can perform tasks, but Toloka chooses instruction language
#   depends on worker language in settings, not language in pool filters. For tasks with language-specific
#   instructions it can cause problems.
def get_task_id(task_spec: PreparedTaskSpec):
    if task_spec.scenario == project.Scenario.DEFAULT:
        id = task_spec.id
    else:
        id = f'{task_spec.id}_{task_spec.scenario.value}'
    return f'{id}_{task_spec.lang}'


def _log_found_project(project_id: str, task_spec: PreparedTaskSpec):
    logger.debug(f'suitable Toloka project {project_id} is found for {task_spec.log_message()}')


def define_task(task_spec: PreparedTaskSpec, client: toloka.TolokaClient, interactive: bool = True):
    if task_spec.scenario == project.Scenario.DEFAULT and isinstance(task_spec, AnnotationTaskSpec):
        define_task(task_spec.check, client, interactive)

    found_project = find_project(task_spec, client)
    if found_project is None:
        prj = toloka.Project(private_comment=get_task_id(task_spec))
    else:
        prj, diff = found_project
        if not diff:
            _log_found_project(prj.id, task_spec)
            logger.info('no changes in task')
            return
        if interactive and not ask(f'{diff}update Toloka project {prj.id} according to the changes shown above?'):
            logger.info('task update is canceled')
            return

    task_spec.task_mapping.validate()

    prj.public_name = task_spec.name[task_spec.lang]
    prj.public_description = task_spec.description[task_spec.lang]
    prj.public_instructions = get_instruction(task_spec)
    prj.task_spec = toloka.project.task_spec.TaskSpec(
        input_spec=task_spec.task_mapping.to_toloka_input_spec(),
        output_spec=task_spec.task_mapping.to_toloka_output_spec(),
        view_spec=task_spec.view,
    )

    if found_project is None:
        logger.debug(f'creating Toloka project for {task_spec.log_message()}')
        client.create_project(prj)
        logger.info('task is defined')
    else:
        logger.debug(f'updating Toloka project {prj.id} for {task_spec.log_message()}')
        client.update_project(prj.id, prj)
        logger.info('task is updated')
        if interactive:
            notify_workers_about_task_change(task_spec, prj, client)


def find_project(
    task_spec: PreparedTaskSpec,
    client: toloka.TolokaClient,
) -> Optional[Tuple[toloka.Project, str]]:
    exists_msg = f'Toloka project is exists for {task_spec.log_message()}'

    for prj in client.get_projects(status='ACTIVE'):
        if not prj.private_comment == get_task_id(task_spec):
            continue
        if not mapping.check_project_is_suitable(prj, task_spec.task_mapping):
            raise ValueError(
                f'{exists_msg}, but function of task is changed in incompatible way. Use different task ID'
            )
        diff = get_project_diff(prj, task_spec)
        diff_msg = f'{exists_msg} and differs from current task spec:\n\n{diff}' if diff else ''
        return prj, diff_msg
    return None


def get_project_diff(prj: toloka.Project, task_spec: PreparedTaskSpec) -> str:
    assert prj.private_comment == get_task_id(task_spec)
    diffs = []
    for current_value, existing_value, value_display, value_name in (
        (task_spec.name[task_spec.lang], prj.public_name, None, 'name'),
        (task_spec.description[task_spec.lang], prj.public_description, None, 'description'),
        (get_instruction(task_spec), prettify_html(prj.public_instructions), None, 'instruction'),
        (
            task_spec.view.config.to_json(pretty=True),
            prj.task_spec.view_spec.config.to_json(pretty=True),
            lambda config: config.to_json(pretty=True),
            'view',
        ),
    ):
        if current_value != existing_value:
            diffs.append(f'{value_name} changed:\n\n{string_diff(existing_value, current_value)}\n\n')
    return '\n'.join(diffs)


def check_project(task_spec: PreparedTaskSpec, client: toloka.TolokaClient) -> toloka.Project:
    found_project = find_project(task_spec, client)
    if found_project is None:
        raise RuntimeError(f'task "{task_spec.id}" is not defined')
    prj, diff = found_project
    if diff:
        raise RuntimeError(diff)
    _log_found_project(prj.id, task_spec)
    return prj


# Toloka supports not all HTML tags and also changes HTML while creating/updating project, using escaping and some
# strange formatting. So we try so compare prettified versions of current and Toloka project instructions. It may break
# if Toloka begins to preprocess HTML in way that our prettifier could not format instructions in the same way. In such
# situation, we will have to put current instruction or its hash to some field in Toloka project, i.e. to private
# comment or metadata, and do not compare with instruction directly.
def prettify_html(html: str) -> str:
    return BeautifulSoup(html, 'html.parser').prettify()


def string_diff(s1: str, s2: str) -> str:
    diff = difflib.unified_diff(s1.splitlines(), s2.splitlines(), lineterm='')
    return '\n'.join(diff_line.rstrip() for diff_line in diff)


def get_instruction(task_spec: PreparedTaskSpec) -> str:
    lang = task_spec.lang
    worker = '<div>' + task_spec.instruction[lang] + '</div>'
    acceptance = instruction.acceptance[lang]

    if isinstance(task_spec.function, base.AnnotationFunction) and not isinstance(task_spec, AnnotationTaskSpec):
        worker = '<div>' + instruction.annotation_check_cut[lang] + worker + '</div>'

    if task_spec.scenario == project.Scenario.DEFAULT:
        return prettify_html(instruction.css + worker + acceptance)

    worker_header, expert_cut = instruction.worker_header[lang], instruction.expert_cut[lang]

    has_evaluation = False
    if isinstance(task_spec.function, base.AnnotationFunction):
        has_evaluation = task_spec.scenario == project.Scenario.EXPERT_LABELING_OF_TASKS

    expert = instruction.get_formatted_expert_instruction(task_spec.function, lang, has_evaluation)

    return prettify_html(instruction.css + worker_header + worker + '<div>' + expert_cut + expert + '</div>')


def notify_workers_about_task_change(task_spec: PreparedTaskSpec, prj: toloka.Project, client: toloka.TolokaClient):
    recipients_filter, recipients_msg = define_notification_recipients(task_spec, prj, client)
    if recipients_filter is None:
        return

    if not ask(f'notify workers of changes in the task? ({recipients_msg})'):
        return

    # If the name of the task has changed, the topic will contain a new name, which may confuse the workers.
    # In this case, we recommend writing in the message about the name change and specify the old name.
    topic = (
        base.LocalizedString(
            {
                'EN': 'Changes in task',
                'RU': 'Изменения в задании',
                'TR': 'Görevlerdeki değişiklikler',
                'KK': 'Тапсырмалардағы өзгерістер',
            }
        )
        + ' "'
        + task_spec.name
        + '"'
    )[task_spec.lang]

    message_lang = f'in {task_spec.lang}'
    if task_spec.lang != base.DEFAULT_LANG:
        message_lang += f' or {base.DEFAULT_LANG}'
    message = input(f'specify message for workers ({message_lang}): ')

    send_message_to_workers(topic, message, task_spec.lang, recipients_filter, client)

    logger.info('workers are notified')


def define_notification_recipients(
    task_spec: PreparedTaskSpec,
    prj: toloka.Project,
    client: toloka.TolokaClient,
) -> Tuple[Optional[toloka.filter.FilterCondition], str]:
    if task_spec.scenario != project.Scenario.DEFAULT:
        skills = experts.get_skills(task_spec.id, task_spec.lang, client)
        if len(skills) == 0:
            logger.info('no registered experts to receive notification')
            return None, ''
        worker_filters = [
            toloka.filter.Skill(key=skill.id, operator=toloka.primitives.operators.CompareOperator.GTE, value=0.0)
            for _, skill in skills
        ]
        recipients_msg = (
            f'all experts who registered for task "{task_spec.id}" in {task_spec.lang} language '
            f'will receive a notification'
        )
    else:
        worker_filters = [toloka.filter.Languages.in_(task_spec.lang)]
        training_requirement = experts.find_training(prj.id, client)
        recipients_msg = (
            f'all {task_spec.lang} workers who have ever performed tasks associated with your Toloka account '
            f'will receive a notification'
        )
        if training_requirement:
            skill_id = experts.get_skill_id_from_training(client.get_pool(training_requirement.id))
            # user skill for training is deleted after retry_training_after_days
            worker_filters.append(
                toloka.filter.Skill(key=skill_id, operator=toloka.primitives.operators.CompareOperator.GTE, value=0.0)
            )
            recipients_msg = (
                f'all workers who have performed the task "{task_spec.id}" in {task_spec.lang} language within '
                f'the last {training_requirement.retry_training_after_days} days will receive a notification'
            )
    return toloka.filter.FilterAnd(worker_filters), recipients_msg


def send_message_to_workers(
    topic: str, message: str, lang: str, recipients_filter: toloka.filter.FilterCondition, client: toloka.TolokaClient
):
    try:
        logger.debug(f'sending message\n\ttopic: {topic}\n\ttext: {message}\n\trecipients filter: {recipients_filter}')
        client.compose_message_thread(
            topic={lang: topic},
            text={lang: message},
            recipients_select_type=toloka.message_thread.RecipientsSelectType.FILTER,
            recipients_filter=recipients_filter,
            answerable=False,
        )
    except toloka.exceptions.IncorrectActionsApiError as e:
        if e.code == 'INCORRECT_RECIPIENTS':
            logger.warn('no recipients for the notification')
        else:
            raise



from dataclasses import dataclass
from typing import List, Tuple, Optional, Dict

from lzy.api.v1 import op
from lzy.api.v1.whiteboard import whiteboard, view
import toloka.client as toloka

from .. import base, client, classification, classification_loop, mapping, pool as pool_config


@dataclass
class Results:
    raw_results: List[Tuple[Optional[classification.TaskLabelsProbas], List[classification.WorkerLabel]]]
    worker_weights: Optional[Dict[str, float]]


@whiteboard(tags=['classification'], namespace='default')
@dataclass
class Whiteboard:
    task_spec: base.TaskSpec
    params: client.params.Params
    project: toloka.Project
    lang: str
    input_objects: List[mapping.Objects]
    control_objects: List[mapping.TaskSingleSolution]

    pool: toloka.Pool = None  # TODO: need to refresh it after end of process, to get actual values for stopped_at
    results: Results = None

    @view
    def to_results(self) -> Results:
        return self.results


@op
def create_pool(
    loop: classification_loop.ClassificationLoop,
    control_objects: List[mapping.TaskSingleSolution],
    pool_cfg: pool_config.ClassificationConfig,
) -> toloka.Pool:
    return loop.create_pool(control_objects, pool_cfg)


@op
def add_input_objects(
    loop: classification_loop.ClassificationLoop,
    input_objects: List[mapping.Objects],
    pool: toloka.Pool,
) -> None:
    loop.add_input_objects(pool.id, input_objects)


@op
def run_loop(
    loop: classification_loop.ClassificationLoop,
    pool: toloka.Pool,
) -> None:
    loop.loop(pool.id)


@op
def get_results(
    loop: classification_loop.ClassificationLoop,
    input_objects: List[mapping.Objects],
    pool: toloka.Pool,
) -> Results:
    result = loop.get_results(pool.id, input_objects)
    return Results(result[0], result[1])



from dataclasses import dataclass
from datetime import timedelta
from typing import Optional

from .. import (
    duration,
    classification,
    classification_loop,
    control as control_params,
    evaluation,
    feedback_loop,
    pricing,
    worker,
)


@dataclass(init=False)
class ExpertParams:
    task_duration_hint: timedelta
    assignment_price: float

    real_tasks_count: int

    @property
    def pricing_config(self) -> pricing.PoolPricingConfig:
        return pricing.PoolPricingConfig(
            assignment_price=self.assignment_price,
            real_tasks_count=self.real_tasks_count,
            control_tasks_count=0,
        )

    def __init__(self, task_duration_hint: timedelta, pricing_config: pricing.PoolPricingConfig):
        self.task_duration_hint = task_duration_hint
        self.assignment_price = pricing_config.assignment_price
        self.real_tasks_count = pricing_config.real_tasks_count


@dataclass(init=False)
class Params(ExpertParams):
    worker_filter: worker.BaseWorkerFilter

    control_tasks_count: int
    pricing_strategy: pricing.PricingStrategy

    overlap: classification_loop.Overlap
    aggregation_algorithm: Optional[classification.AggregationAlgorithm]

    control: control_params.Control

    # TODO(DATAFORGE-75): model is always applied on first loop iteration
    model: Optional[worker.Model]

    task_duration_function: Optional[duration.TaskDurationFunction]

    def __init__(
        self,
        task_duration_hint: timedelta,
        pricing_config: pricing.PoolPricingConfig,
        overlap: classification_loop.Overlap,
        control: control_params.Control,
        worker_filter: worker.BaseWorkerFilter,
        aggregation_algorithm: Optional[classification.AggregationAlgorithm] = None,
        pricing_strategy: pricing.PricingStrategy = pricing.StaticPricingStrategy(),
        task_duration_function: Optional[duration.TaskDurationFunction] = None,
    ):
        super(Params, self).__init__(task_duration_hint, pricing_config)
        self.control_tasks_count = pricing_config.control_tasks_count
        self.overlap = overlap
        self.aggregation_algorithm = aggregation_algorithm
        self.control = control
        self.worker_filter = worker_filter
        self.pricing_strategy = pricing_strategy
        self.model = None
        self.task_duration_function = task_duration_function

    @property
    def pricing_config(self) -> pricing.PoolPricingConfig:
        return pricing.PoolPricingConfig(
            assignment_price=self.assignment_price,
            real_tasks_count=self.real_tasks_count,
            control_tasks_count=self.control_tasks_count,
        )

    @property
    def classification_loop_params(self) -> classification_loop.Params:
        task_duration_function = self.task_duration_function
        if task_duration_function is None:
            task_duration_function = duration.get_const_task_duration_function(self.task_duration_hint)
        return classification_loop.Params(
            aggregation_algorithm=self.aggregation_algorithm,
            overlap=self.overlap,
            control=self.control,
            task_duration_function=task_duration_function,
        )


@dataclass(init=False)
class AnnotationParams(Params):
    assignment_check_sample: Optional[evaluation.AssignmentCheckSample]
    overlap: classification_loop.DynamicOverlap

    def __init__(
        self,
        task_duration_hint: timedelta,
        pricing_config: pricing.PoolPricingConfig,
        overlap: classification_loop.Overlap,
        control: control_params.Control,
        worker_filter: worker.BaseWorkerFilter,
        aggregation_algorithm: Optional[classification.AggregationAlgorithm] = None,
        pricing_strategy: pricing.PricingStrategy = pricing.StaticPricingStrategy(),
        assignment_check_sample: Optional[evaluation.AssignmentCheckSample] = None,
        task_duration_function: Optional[duration.TaskDurationFunction] = None,
    ):
        super(AnnotationParams, self).__init__(
            task_duration_hint,
            pricing_config,
            overlap,
            control,
            worker_filter,
            aggregation_algorithm,
            pricing_strategy,
            task_duration_function,
        )
        assert isinstance(
            self.overlap, classification_loop.DynamicOverlap
        ), 'You should use dynamic overlap for annotation'
        assert self.overlap.min_overlap == 1, 'For annotation we always start with one attempt'
        assert self.aggregation_algorithm is None, 'No aggregation algorithm should be specified for annotation params'
        self.assignment_check_sample = assignment_check_sample
        assert (
            self.pricing_config.control_tasks_count == 0
        ), 'We cannot use control tasks as measure of quality control for annotation tasks'

    @property
    def feedback_loop_params(self) -> feedback_loop.Params:
        task_duration_function = self.task_duration_function
        if task_duration_function is None:
            task_duration_function = duration.get_const_task_duration_function(self.task_duration_hint)
        return feedback_loop.Params(self.assignment_check_sample, self.control, self.overlap, task_duration_function)



import abc
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import List, Dict, Any

from ..precision import FLOAT_PRECISION_PLACES
from .common import DEFAULT_LANG, LocalizedString


class Object:
    @staticmethod
    def is_media() -> bool:
        return False


class Label:
    @staticmethod
    @abc.abstractmethod
    def possible_values() -> list:
        ...

    @staticmethod
    @abc.abstractmethod
    def possible_instances() -> List['Label']:
        ...

    @staticmethod
    @abc.abstractmethod
    def get_display_labels() -> List[LocalizedString]:
        ...


class Class(Label, Object, Enum):
    @classmethod
    def possible_values(cls) -> List[str]:
        return [i.value for i in cls.possible_instances()]

    @classmethod
    def possible_instances(cls) -> List['Class']:
        return list(cls)

    @classmethod
    def get_display_labels(cls) -> List[LocalizedString]:
        return [i.get_label() for i in cls.possible_instances()]

    def get_label(self) -> LocalizedString:
        return LocalizedString(type(self).labels()[self])

    @classmethod
    def labels(cls) -> Dict['Class', Dict[str, str]]:
        return {value: {DEFAULT_LANG: value.value} for value in cls.possible_instances()}

    def __lt__(self, other):
        if isinstance(other, Class):
            return self.value < other.value
        # during crowd-kit Dawid-Skene aggregation values are sometimes compared to column names
        assert isinstance(other, str)
        return self.value < other

    def __new__(cls, *args):
        # todo: subclass Class from StrEnum, when it becomes available
        value = args[0]
        assert isinstance(value, str), 'only str-valued Classes are accepted'
        obj = object.__new__(cls)
        obj._value_ = value
        return obj


class SbSChoice(Class):
    A = 'a'
    B = 'b'

    @classmethod
    def labels(cls) -> Dict['SbSChoice', Dict[str, str]]:
        return {
            cls.A: {
                'EN': 'Left option',
                'RU': 'Левый вариант',
                'TR': 'Sol seçenek',
                'KK': 'Сол жақтағы опция',
            },
            cls.B: {
                'EN': 'Right option',
                'RU': 'Правый вариант',
                'TR': 'Doğru seçenek',
                'KK': 'Оң жақтағы опция',
            },
        }

    def swap(self) -> 'SbSChoice':
        if self == self.A:
            return self.B
        return self.A


class Evaluation(Object):
    ...


@dataclass(eq=True, order=True, frozen=True)
class BinaryEvaluation(Evaluation, Label):
    ok: bool

    @staticmethod
    def possible_values() -> List[bool]:
        return [True, False]

    @staticmethod
    def possible_instances() -> List['BinaryEvaluation']:
        return [BinaryEvaluation(ok=value) for value in BinaryEvaluation.possible_values()]

    @staticmethod
    def get_display_labels() -> List[LocalizedString]:
        return [
            LocalizedString({'EN': 'Yes', 'RU': 'Да', 'TR': 'Evet', 'KK': 'Иә'}),
            LocalizedString({'EN': 'No', 'RU': 'Нет', 'TR': 'Hayır', 'KK': 'Жоқ'}),
        ]


class ScoreEvaluation(Evaluation, Class):
    ...


@dataclass(frozen=True)
class Metadata(Object):
    metadata: str


def _change_type(
    data: Any,
    from_type: type,
    to_type: type,
    change_precision: bool = False,
) -> Any:
    if isinstance(data, dict):
        return {key: _change_type(value, from_type, to_type, change_precision) for key, value in data.items()}
    elif isinstance(data, list):
        return [_change_type(item, from_type, to_type, change_precision) for item in data]
    elif isinstance(data, from_type):
        data = to_type(data)
    if isinstance(data, to_type) and change_precision:
        data = round(data, FLOAT_PRECISION_PLACES)
    return data


@dataclass(init=False, frozen=True)
class ImageAnnotation(Object):
    data: list

    def __init__(self, data: list):
        object.__setattr__(self, 'data', _change_type(data, float, Decimal))

    def __repr__(self) -> str:
        return f"{type(self).__name__}(data={_change_type(self.data, float, Decimal, change_precision=True)})"

    def __hash__(self) -> int:
        return hash(repr(self))



import abc
from dataclasses import dataclass
import re
from typing import Tuple, Union, Type, Optional

import pandas as pd

from .common import LocalizedString, Title
from .metas import ClassMeta, EvaluationMeta, ObjectMeta, SbSChoiceMeta
from .types import BinaryEvaluation, Class, Evaluation, Object, SbSChoice, ImageAnnotation

FunctionArgument = Union[Type[Object], ObjectMeta]
FunctionArguments = Tuple[FunctionArgument, ...]

CLASS_TASK_FIELD = 'choice'
CLASS_OBJ_FIELD = 'value'
EVALUATION_TASK_FIELD = 'eval'


class TaskFunction:
    def __post_init__(self):
        self.has_names = None
        self.has_names = self.has_names_in_metas(self.get_all_arguments())
        self.additional_validate()
        self.validate(self.get_all_arguments())

    @staticmethod
    def has_names_in_metas(metas: Tuple[ObjectMeta, ...]) -> bool:
        return any(meta.name is not None for meta in metas)

    @abc.abstractmethod
    def get_inputs(self) -> Tuple[ObjectMeta, ...]:
        ...

    @abc.abstractmethod
    def get_outputs(self) -> Tuple[ObjectMeta, ...]:
        ...

    def get_all_arguments(self) -> Tuple[ObjectMeta, ...]:
        return self.get_inputs() + self.get_outputs()

    @staticmethod
    def display_arguments(metas: Tuple[ObjectMeta, ...]) -> str:
        return str(pd.DataFrame([{'type': meta.type.__name__, 'name': meta.name} for meta in metas]))

    def print_arguments(self):
        print(self.display_arguments(self.get_all_arguments()))

    def additional_validate(self):
        pass

    @staticmethod
    def validate(metas: Tuple[ObjectMeta, ...]):
        has_name = TaskFunction.has_names_in_metas(metas)
        if has_name:
            names = [meta.name for meta in metas]
            classes_repr = f'Current names:\n{TaskFunction.display_arguments(metas)}'
            assert all(name is not None for name in names), f'You should provide names for all types. {classes_repr}'
            assert len(names) == len(set(names)), f'Names should be unique. {classes_repr}'
            pattern = re.compile(r'\w{1,30}', re.ASCII)
            for name in names:
                assert not name.startswith('_'), 'Names starting with "_" are reserved for internal usage'
                result = pattern.fullmatch(name)
                assert result is not None, f'Incorrect name: "{name}"'

    @staticmethod
    def convert_to_meta(type_or_meta: FunctionArgument, name: Optional[str] = None) -> ObjectMeta:
        if isinstance(type_or_meta, ObjectMeta):
            return type_or_meta
        elif isinstance(type_or_meta, type) and issubclass(type_or_meta, Object):
            # todo name?
            return ObjectMeta[type_or_meta](type=type_or_meta, name=name)
        raise ValueError(f'Unexpected type or instance: "{type_or_meta}"')

    @staticmethod
    def convert_to_metas(types: Tuple[FunctionArgument]) -> Tuple[ObjectMeta, ...]:
        return tuple(TaskFunction.convert_to_meta(type_or_meta) for type_or_meta in types)


@dataclass
class ClassificationFunction(TaskFunction):
    inputs: FunctionArguments
    cls: Union[Type[Class], ClassMeta]

    def get_inputs(self) -> Tuple[ObjectMeta, ...]:
        return self.convert_to_metas(self.inputs)

    def get_outputs(self) -> Tuple[ObjectMeta, ...]:
        return (self.convert_to_meta(self.cls, name=CLASS_TASK_FIELD if self.has_names else None),)

    def additional_validate(self):
        assert (
            isinstance(self.cls, ClassMeta)
            and issubclass(self.cls.type, Class)
            or isinstance(self.cls, type)
            and issubclass(self.cls, Class)
        )
        assert not isinstance(self.cls, ObjectMeta) or self.cls.required, 'choice is required'


@dataclass
class SbSFunction(TaskFunction):
    inputs: FunctionArguments
    hints: Optional[FunctionArguments] = None
    choice: Union[Type[SbSChoice], SbSChoiceMeta] = SbSChoice

    def get_inputs(self) -> Tuple[ObjectMeta, ...]:
        return self.convert_to_metas(self.inputs)

    def get_outputs(self) -> Tuple[ObjectMeta, ...]:
        return (self.convert_to_meta(self.choice, name=CLASS_TASK_FIELD if self.has_names else None),)

    def get_hints(self) -> Tuple[ObjectMeta, ...]:
        return self.convert_to_metas(self.hints or ())

    def get_all_arguments(self) -> Tuple[ObjectMeta, ...]:
        return self.get_hints() + super(SbSFunction, self).get_all_arguments()

    def additional_validate(self):
        if not (isinstance(self.choice, type) and issubclass(self.choice, SbSChoice)):
            raise NotImplementedError(f'Unexpected type or instance: "{self.choice}"')
        assert not isinstance(self.choice, ObjectMeta) or self.choice.required, 'choice is required'


@dataclass
class AnnotationFunction(TaskFunction):
    inputs: FunctionArguments
    outputs: FunctionArguments
    evaluation: Type[Evaluation] = BinaryEvaluation

    def has_image_annotation(self) -> bool:
        return any(output_meta.type == ImageAnnotation for output_meta in self.get_outputs())

    def get_inputs(self) -> Tuple[ObjectMeta, ...]:
        return self.convert_to_metas(self.inputs)

    def get_outputs(self) -> Tuple[ObjectMeta, ...]:
        return self.convert_to_metas(self.outputs)

    def get_evaluation(self) -> EvaluationMeta:
        return EvaluationMeta(
            type=self.evaluation,
            name=EVALUATION_TASK_FIELD if self.has_names else None,
            title=Title(
                text=LocalizedString(
                    {
                        'EN': 'Is the solution accurate?',
                        'RU': 'Решение верное?',
                        'TR': 'Görevin kararı doğru mu?',
                        'KK': 'Тапсырманың шешімі дұрыс па?',
                    }
                )
            ),
        )

    def get_all_arguments(self) -> Tuple[ObjectMeta, ...]:
        return super(AnnotationFunction, self).get_all_arguments() + (self.get_evaluation(),)

    def additional_validate(self):
        if not (isinstance(self.evaluation, type) and issubclass(self.evaluation, Evaluation)):
            raise NotImplementedError(f'Unexpected type: "{self.evaluation}"')

    def get_evaluation_named_meta(self) -> EvaluationMeta:
        # todo: may be incorrect sometimes
        evaluation = self.get_evaluation()
        if evaluation.name is None:
            evaluation.name = EVALUATION_TASK_FIELD
        return evaluation



from collections import defaultdict
from contextlib import AbstractContextManager
from dataclasses import dataclass, is_dataclass
from typing import List, Type, Tuple, Optional, Dict, Iterable, Any

import toloka.client as toloka

import dataforge.base as base

obj_primitive_field_types = {str, int, bool, list}

validation_enabled = True


class DisableValidation(AbstractContextManager):
    def __enter__(self):
        global validation_enabled
        validation_enabled = False

    def __exit__(self, exc_type, exc, exc_tb):
        global validation_enabled
        validation_enabled = True


def obj_to_values(obj: base.Object) -> dict:
    if issubclass(type(obj), base.Class):
        assert isinstance(obj.value, str), 'class enum must have string values'
        return {base.CLASS_OBJ_FIELD: obj.value}
    values = {}
    for field_name, field_value in obj.__dict__.items():
        if any(isinstance(field_value, t) for t in obj_primitive_field_types):
            values[field_name] = field_value
        else:
            raise ValueError(f'object field "{field_name}" has unsupported type {type(field_value)}')
    return values


def values_to_obj(values: dict, obj_cls: Type[base.Object]) -> base.Object:
    if issubclass(obj_cls, base.Class):
        return obj_cls(values[base.CLASS_OBJ_FIELD])
    fields = {}
    for field_name, field_value in values.items():
        field_type = _get_all_annotations(obj_cls).get(field_name)
        if field_type is None:
            continue  # additional toloka values is OK
        if field_type in obj_primitive_field_types:
            assert field_type == type(field_value), (
                f'obj field type {field_type} does not match to values type ' f'{type(field_value)}'
            )
            fields[field_name] = field_value
        else:
            raise ValueError(f'value {field_name} has unsupported type {type(field_value)}')
    return obj_cls(**fields)


def _get_all_annotations(obj_cls: Type[base.Object]) -> dict:
    """
    Get all class annotations, including annotations from superclasses.
    In case of annotations collision, superclass annotations will be overridden.
    """
    result = {}
    for c in obj_cls.mro()[::-1]:
        if hasattr(c, '__annotations__'):
            result.update(c.__annotations__)
    if len(result) == 0:
        raise AttributeError('class must have annotations for serialization')
    return result


@dataclass(frozen=True)
class ObjectMapping:
    obj_meta: base.ObjectMeta
    obj_task_fields: Tuple[Tuple[str, str], ...]

    @property
    def obj_type(self) -> base.ObjectT:
        return self.obj_meta.type

    def to_values(self, obj: Optional[base.Object]) -> Dict[str, Any]:
        if validation_enabled:
            if obj is None:
                assert not self.obj_meta.required, f'object {self.obj_meta} is required'
            else:
                assert (
                    obj.__class__ == self.obj_type
                ), f'passed {obj.__class__} does not correspond to mapping {self.obj_type}'
        if obj is None:
            return {}
        obj_values = obj_to_values(obj)
        values = {}
        for obj_field, task_field in self.obj_task_fields:
            values[task_field] = obj_values[obj_field]
        return values

    def to_toloka_spec(self) -> Dict[str, toloka.primitives.base.BaseTolokaObject]:
        type_to_toloka_spec = {
            str: toloka.project.StringSpec,
            bool: toloka.project.BooleanSpec,
            int: toloka.project.IntegerSpec,
            float: toloka.project.FloatSpec,
            list: toloka.project.JsonSpec,
        }

        if issubclass(self.obj_type, base.Label):
            assert len(self.obj_task_fields) == 1
            allowed_values = list(sorted(self.obj_type.possible_values()))
            assert len(allowed_values) > 0
            return {
                self.obj_task_fields[0][1]: type_to_toloka_spec[type(allowed_values[0])](allowed_values=allowed_values)
            }

        spec = {}
        for obj_field, task_field in self.obj_task_fields:
            obj_field_type = self.obj_type.__annotations__[obj_field]
            if obj_field_type == str and self.obj_type.is_media():
                toloka_spec_type = toloka.project.UrlSpec
            else:
                toloka_spec_type = type_to_toloka_spec[obj_field_type]
            spec[task_field] = toloka_spec_type(
                required=self.obj_meta.required, hidden=issubclass(self.obj_type, base.Metadata)
            )
        return spec

    def from_values(self, values: Dict[str, Any]) -> Optional[base.Object]:
        task_fields = set(task_field for _, task_field in self.obj_task_fields)
        if task_fields & values.keys() != task_fields:
            assert (
                not self.obj_meta.required
            ), f'missing fields {task_fields - values.keys()} in required object {self.obj_meta}'
            return None
        return values_to_obj(
            values={obj_field: values[task_field] for obj_field, task_field in self.obj_task_fields},
            obj_cls=self.obj_type,
        )

    def validate(self):
        if is_dataclass(self.obj_type):
            for obj_field, _ in self.obj_task_fields:
                assert (
                    obj_field in self.obj_type.__annotations__
                ), f'{self.obj_type} does not have attribute "{obj_field}"'
        elif issubclass(self.obj_type, base.Class):
            # assert self.obj_task_fields == ((base.CLASS_OBJ_FIELD, base.CLASS_TASK_FIELD),)
            # variant above is stricter, but we have some deps, which use self-defined mapping
            assert len(self.obj_task_fields) == 1 and self.obj_task_fields[0][0] == base.CLASS_OBJ_FIELD
        else:
            assert False, f'{self.obj_type} is not a dataclass nor an options class'


TASK_ID_FIELD = 'id'
Objects = Tuple[Optional[base.Object], ...]


class TaskID:
    objects: Objects
    id: str

    def __init__(self, objects: Objects):
        self.objects = objects
        # Нужно как-то сопоставлять объекты задания с соответствующим toloka.Task при выгрузке
        # страниц заданий. Можно было бы назначать каждому объекту случайный ID, но тогда некоторые
        # операции, вроде feedback loop, потеряют реентерабельность. Этот подход можно использовать
        # только при поддержке реентерабельности со стороны среды исполнения (возможно будет в lzy).
        # Пока этого нет, рассчитываем на реализацию __repr__() @dataclass-ов, который сериализует
        # фактические данные объекта в строку. С float-ми засада с числом знаков, но мы обязаны
        # воспроизводить результаты при перезапуске на том же env, что перед падением, так что ОК.
        self.id = ' '.join(repr(obj) for obj in objects)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __repr__(self) -> str:
        return self.id


TaskSingleSolution = Tuple[Objects, Objects]


@dataclass(frozen=True)
class TaskMapping:
    input_mapping: Tuple[ObjectMapping, ...]
    output_mapping: Tuple[ObjectMapping, ...]

    def to_task(self, objects: Objects) -> toloka.Task:
        if validation_enabled:
            self.validate_objects(objects)
        return toloka.Task(input_values=self.toloka_values(objects))

    def to_control_task(self, control_objects: TaskSingleSolution) -> toloka.Task:
        input_objects, output_objects = control_objects
        task = self.to_task(input_objects)
        self.validate_objects(output_objects, output=True)
        task.known_solutions = [
            toloka.Task.KnownSolution(
                output_values=self.toloka_values(output_objects, output=True), correctness_weight=1
            )
        ]
        task.infinite_overlap = True
        return task

    def to_solution(self, objects: Objects) -> toloka.solution.Solution:
        # mostly for tests, because solutions are created in Toloka usually
        self.validate_objects(objects, output=True)
        return toloka.solution.Solution(output_values=self.toloka_values(objects, output=True))

    @staticmethod
    def objects_mapping_to_toloka_spec(
        mapping: Tuple[ObjectMapping, ...],
    ) -> Dict[str, toloka.primitives.base.BaseTolokaObject]:
        spec = {}
        for object_mapping in mapping:
            object_spec = object_mapping.to_toloka_spec()
            assert not spec.keys() & object_spec
            spec.update(object_spec)
        return spec

    def to_toloka_input_spec(self) -> Dict[str, toloka.primitives.base.BaseTolokaObject]:
        spec = self.objects_mapping_to_toloka_spec(self.input_mapping)
        assert TASK_ID_FIELD not in spec
        spec[TASK_ID_FIELD] = toloka.project.StringSpec(hidden=True)
        return spec

    def to_toloka_output_spec(self) -> Dict[str, toloka.primitives.base.BaseTolokaObject]:
        return self.objects_mapping_to_toloka_spec(self.output_mapping)

    def toloka_values(self, objects: Objects, output: bool = False) -> Dict[str, Any]:
        mapping = self.output_mapping if output else self.input_mapping
        values = {}
        for obj, obj_mapping in zip(objects, mapping):
            values.update(obj_mapping.to_values(obj))
        if not output:
            values[TASK_ID_FIELD] = self.task_id(objects).id
        return values

    def task_id(self, objects: Objects) -> TaskID:
        if validation_enabled:
            self.validate_objects(objects)
        return TaskID(objects)

    def from_task(self, task: toloka.Task) -> Objects:
        return self.from_task_values(task.input_values)

    def from_task_values(self, values: Dict[str, Any]) -> Objects:
        return tuple(mapping.from_values(values) for mapping in self.input_mapping)

    def from_solution(self, solution: toloka.solution.Solution) -> Objects:
        return self.from_solution_values(solution.output_values)

    def from_solution_values(self, values: Dict[str, Any]) -> Objects:
        return tuple(mapping.from_values(values) for mapping in self.output_mapping)

    def validate(self):
        self.validate_mapping_list(self.input_mapping)
        self.validate_mapping_list(self.output_mapping)

    @staticmethod
    def validate_mapping_list(mapping_list: Tuple[ObjectMapping, ...]):
        fields = []
        for mapping in mapping_list:
            mapping.validate()
            fields += [task_field for _, task_field in mapping.obj_task_fields]
        assert len(fields) == len(set(fields)), f'fields intersection in mapping list: {fields}'

    def validate_objects(self, objects: Objects, output: bool = False):
        mapping = self.output_mapping if output else self.input_mapping
        assert len(objects) == len(mapping), f'expected length {len(mapping)}, got {len(objects)}'
        for obj, obj_mapping in zip(objects, mapping):
            assert (
                obj_mapping.obj_type == obj.__class__ or not obj_mapping.obj_meta.required and obj is None
            ), f'expected type {obj_mapping.obj_type}, got {obj.__class__}'


TaskMultipleSolutions = Tuple[Objects, List[Tuple[Objects, toloka.Assignment]]]
TaskMultipleSolutionOptions = Tuple[Objects, List[Objects]]
AssignmentSolutions = Tuple[toloka.Assignment, List[TaskSingleSolution]]


# solutions grouped by assignments, assignments order is preserved
def get_assignments_solutions(
    assignments: List[toloka.Assignment],
    mapping: TaskMapping,
    with_control_tasks: bool = False,
) -> List[AssignmentSolutions]:
    assignments_solutions = []
    for assignment in assignments:
        solutions = []
        for task_id, input_objects, output_objects in iterate_assignment(assignment, mapping, with_control_tasks):
            solutions.append((input_objects, output_objects))
        assignments_solutions.append((assignment, solutions))
    return assignments_solutions


# solutions grouped by input objects, input objects order is preserved
def get_solutions(
    assignments: List[toloka.Assignment],
    mapping: TaskMapping,
    pool_input_objects: List[Objects],
) -> List[TaskMultipleSolutions]:
    task_id_to_output_objects = defaultdict(list)
    for assignment in assignments:
        for task_id, _, output_objects in iterate_assignment(assignment, mapping):
            task_id_to_output_objects[task_id].append((output_objects, assignment))
    return [
        (input_objects, task_id_to_output_objects[mapping.task_id(input_objects).id])
        for input_objects in pool_input_objects
    ]


def iterate_assignment(
    assignment: toloka.Assignment,
    mapping: TaskMapping,
    with_control_tasks: bool = False,
) -> Iterable[Tuple[str, Objects, Objects]]:
    return (
        (task.input_values[TASK_ID_FIELD], mapping.from_task(task), mapping.from_solution(solution))
        for task, solution in zip(assignment.tasks, assignment.solutions)
        if (with_control_tasks or not task.known_solutions)
    )


def iterate_assignment_tasks(
    assignment: toloka.Assignment,
    mapping: TaskMapping,
    with_control_tasks: bool = True,
) -> Iterable[Tuple[str, Objects]]:
    return (
        (task.input_values[TASK_ID_FIELD], mapping.from_task(task))
        for task in assignment.tasks
        if (with_control_tasks or not task.known_solutions)
    )


def generate_feedback_loop_mapping(
    input_objects_mapping: Tuple[ObjectMapping, ...],
    markup_objects_mapping: Tuple[ObjectMapping, ...],
    check_objects_mapping: Tuple[ObjectMapping, ...],
) -> Tuple[TaskMapping, TaskMapping]:
    # first object in check output objects must be evaluation object
    assert issubclass(check_objects_mapping[0].obj_type, base.Evaluation)

    for mapping in input_objects_mapping + markup_objects_mapping + check_objects_mapping:
        mapping.validate()

    return TaskMapping(
        input_mapping=input_objects_mapping,
        output_mapping=markup_objects_mapping,
    ), TaskMapping(input_mapping=input_objects_mapping + markup_objects_mapping, output_mapping=check_objects_mapping)


def check_project_is_suitable(project: toloka.Project, task_mapping: TaskMapping) -> bool:
    # toloka specs are unhashable
    def is_included(
        inner: Dict[str, toloka.primitives.base.BaseTolokaObject],
        outer: Dict[str, toloka.primitives.base.BaseTolokaObject],
    ) -> bool:
        # TODO: allow transition from required=True to required=False; the reverse is prohibited
        return all(key in outer and outer[key] == inner[key] for key in inner)

    generated_input_spec = task_mapping.to_toloka_input_spec()
    generated_output_spec = task_mapping.to_toloka_output_spec()

    return is_included(generated_input_spec, project.task_spec.input_spec) and is_included(
        generated_output_spec, project.task_spec.output_spec
    )



from random import choice
import threading
from typing import List, Tuple, Optional, Dict

import toloka.client as toloka

from .. import base, classification, mapping

from .task import PreparedTaskSpec
from .launch import Params, launch
from .relaunch import get_swaps


# In a side-by-side comparison, client has two variants of objects, list of A's and list of B's, and want to know which
# variant, A or B, is better according to his criteria. By default, we place A's to the left, but to avoid workers bias
# of left/right variants of task UI or different exploits, we always randomly swap each pair of variants before creating
# tasks for workers. Overall process looks like:
# - take A's and B's (client input)
# - random swap, create L's and R's for tasks (remember swaps)
# - labeling
# - get results as L's and R's
# - convert L's and R's back to A's and B's (using remembered swaps)
# - return results for each pair in A's and B's terms (client output)
def launch_sbs(
    task_spec: PreparedTaskSpec,
    params: Params,
    input_objects: List[mapping.Objects],
    control_objects: List[mapping.TaskSingleSolution],  # assisted by us
    client: toloka.TolokaClient,
    interactive: bool = False,
    interrupted: Optional[threading.Event] = None,
    name: Optional[str] = None,
) -> Tuple[
    List[Tuple[Optional[classification.TaskLabelsProbas], List[classification.WorkerLabel]]],
    Optional[Dict[str, float]],
]:
    assert isinstance(task_spec.function, base.SbSFunction)

    # [hint, option_a, option_b] = [choice_a | choice_b]
    # swap: option_a <-> option_b, choice_a <-> choice_b

    h_cnt = len(task_spec.function.get_hints())
    i_cnt = len(task_spec.function.get_inputs())
    a_fr = h_cnt
    b_fr = h_cnt + i_cnt
    b_to = h_cnt + i_cnt * 2

    def swap_task(task: mapping.Objects) -> mapping.Objects:
        result = list(task)
        result[a_fr:b_fr], result[b_fr:b_to] = result[b_fr:b_to], result[a_fr:b_fr]
        return tuple(result)

    def swap_task_solution(task_solution: mapping.TaskSingleSolution) -> mapping.TaskSingleSolution:
        task, solution = task_solution
        return swap_task(task), (solution[0].swap(),) + solution[1:]

    input_objects_swapped = []
    swaps = get_swaps(name, input_objects)
    for swap, task_input_objects in zip(swaps, input_objects):
        swaps.append(swap)
        if swap:
            task_input_objects = swap_task(task_input_objects)
        input_objects_swapped.append(task_input_objects)

    control_objects_swapped = []
    for control_object in control_objects:
        swap = choice((True, False))
        if swap:
            control_object = swap_task_solution(control_object)
        control_objects_swapped.append(control_object)

    results_swapped, worker_weights = launch(
        task_spec=task_spec,
        params=params,
        input_objects=input_objects_swapped,
        control_objects=control_objects_swapped,
        client=client,
        interactive=interactive,
        interrupted=interrupted,
        name=name,
    )

    results = []
    for (labels_probas, worker_labels), swap in zip(results_swapped, swaps):
        assert labels_probas is not None
        if swap:
            labels_probas = {label.swap(): proba for label, proba in labels_probas.items()}
            worker_labels = [(label.swap(), worker) for label, worker in worker_labels]
        results.append((labels_probas, worker_labels))
    return results, worker_weights



import logging
from typing import List, Dict, Tuple, Optional
from dataclasses import dataclass

import toloka.client as toloka

from .. import classification_loop, datasource, evaluation, mapping, pool as pool_config, utils, worker

from .params import Params, Evaluation
from .results import Solution, get_results

logger = logging.getLogger(__name__)


@dataclass
class CheckMarkupIterationStats:
    tasks_to_re_markup: int
    ok_tasks_ratio: float


class FeedbackLoop:
    pool_input_objects: List[mapping.Objects]
    check_params: classification_loop.Params
    markup_params: Params
    evaluation: Evaluation
    markup_task_mapping: mapping.TaskMapping
    check_task_mapping: mapping.TaskMapping
    check_loop: classification_loop.ClassificationLoop
    client: toloka.TolokaClient
    lang: str
    interrupter: Optional[utils.LaunchInterrupter]
    s3: Optional[datasource.S3]
    model_ws: Optional[worker.ModelWorkspace]

    def __init__(
        self,
        pool_input_objects: List[mapping.Objects],
        markup_task_mapping: mapping.TaskMapping,
        check_task_mapping: mapping.TaskMapping,
        markup_params: Params,
        check_params: classification_loop.Params,
        client: toloka.TolokaClient,
        lang: str,
        interrupter: Optional[utils.LaunchInterrupter] = None,
        s3: Optional[datasource.S3] = None,
        model_markup: Optional[worker.Model] = None,
        model_check: Optional[worker.Model] = None,
    ):
        self.evaluation = Evaluation(
            aggregation_algorithm=check_params.aggregation_algorithm,
            assignment_check_sample=markup_params.assignment_check_sample,
            overlap=check_params.overlap,
        )

        self.pool_input_objects = pool_input_objects
        self.check_params = check_params
        self.markup_params = markup_params
        self.client = client
        self.markup_task_mapping = markup_task_mapping
        self.check_task_mapping = check_task_mapping
        self.check_loop = classification_loop.ClassificationLoop(
            client=client,
            task_mapping=self.check_task_mapping,
            params=check_params,
            lang=lang,
            interrupter=interrupter,
            model=model_check,
            # TODO
            #  1) also look at control tasks count for check pool
            #  2) (DATAFORGE-75): correct only when model substitutes all solutions
            with_control_tasks=model_check is None,
        )
        self.lang = lang
        self.interrupter = interrupter
        self.s3 = s3
        self.model_ws = None
        if model_markup:
            self.model_ws = worker.ModelWorkspace(model=model_markup, task_mapping=self.markup_task_mapping)

    def create_pools(
        self,
        control_objects: List[mapping.TaskSingleSolution],
        markup_pool_config: pool_config.MarkupConfig,
        check_pool_config: pool_config.ClassificationConfig,
        check_projects: bool = True,  # todo tmp for old pipelines
    ) -> Tuple[toloka.Pool, toloka.Pool]:
        check_sample = self.evaluation.assignment_check_sample
        if check_sample is not None and check_sample.max_tasks_to_check is not None:
            assert check_sample.max_tasks_to_check <= markup_pool_config.real_tasks_count
        if check_projects:
            assert mapping.check_project_is_suitable(
                self.client.get_project(markup_pool_config.project_id), self.markup_task_mapping
            )

        markup_pool = self.client.create_pool(pool_config.create_pool_params(markup_pool_config))

        markup_tasks = []
        for input_objects in self.pool_input_objects:
            task = self.markup_task_mapping.to_task(input_objects)
            task.pool_id = markup_pool.id
            markup_tasks.append(task)

        logger.debug(f'creating {len(markup_tasks)} markup tasks')
        self.client.create_tasks(markup_tasks, allow_defaults=True, async_mode=True, skip_invalid_items=False)

        check_pool = self.check_loop.create_pool(control_objects, check_pool_config, check_projects)

        if self.model_ws is None:
            # If model is presented, first attempt is by it, no need to open pool.
            # In case of rejected model solutions pool will be opened later.
            self.client.open_pool(markup_pool.id)

        return markup_pool, check_pool

    def loop(self, markup_pool_id: str, check_pool_id: str):
        iteration = 1
        while True:
            logger.debug(f'feedback loop iteration #{iteration} is started')
            markup_assignments, fast_markup_assignments = self.get_markups(markup_pool_id, check_pool_id)
            check_stats = self.check_markups(markup_pool_id, check_pool_id, markup_assignments, fast_markup_assignments)
            if check_stats is None:
                break
            iteration += 1
        self.give_bonuses_to_users(check_pool_id=check_pool_id, markup_assignments=markup_assignments)

    def give_bonuses_to_users(
        self,
        check_pool_id: str,
        markup_assignments: List[mapping.AssignmentSolutions],
    ) -> Optional[toloka.operations.Operation]:
        bonuses = evaluation.calculate_bonuses_for_pool_of_markup_assignments(
            markup_assignments=markup_assignments,
            solution_id_to_evaluation=self.get_checks(check_pool_id),
            check_task_mapping=self.check_task_mapping,
            control_params=self.markup_params.control,
        )
        if not bonuses:
            return None

        logger.debug(f'giving {len(bonuses)} bonuses to users')
        op = self.client.create_user_bonuses_async(bonuses)
        result = self.client.wait_operation(op)
        assert (
            result.status == toloka.operations.Operation.Status.SUCCESS
        ), f'bonus issuing failed, operation id: {result.id}, details: {result.details}'

        return result

    def get_results(
        self,
        markup_pool_id: str,
        check_pool_id: str,
    ) -> Tuple[List[List[Solution]], Optional[Dict[str, float]]]:
        checks, worker_weights = self.get_checks_and_weights(check_pool_id)
        return (
            get_results(
                self.pool_input_objects,
                self.get_markups(markup_pool_id, check_pool_id)[0],
                checks,
                self.markup_task_mapping,
                self.check_task_mapping,
            ),
            worker_weights,
        )

    def get_markups(
        self, markup_pool_id: str, check_pool_id: str
    ) -> Tuple[List[mapping.AssignmentSolutions], List[mapping.AssignmentSolutions]]:
        filtered_human_markups, fast_markups = evaluation.prior_filter_assignments(
            self.client,
            self.get_human_markups(markup_pool_id),
            self.markup_params.control,
            self.lang,
            task_duration_function=self.markup_params.task_duration_function,
        )
        return filtered_human_markups + self.get_model_markups(check_pool_id), fast_markups

    def get_human_markups(self, pool_id: str) -> List[mapping.AssignmentSolutions]:
        utils.wait_pool_for_close(self.client, pool_id, self.interrupter)
        return datasource.substitute_media_output(
            mapping.get_assignments_solutions(
                assignments=list(
                    self.client.get_assignments(
                        status=[toloka.Assignment.SUBMITTED, toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED],
                        pool_id=pool_id,
                    )
                ),
                mapping=self.markup_task_mapping,
            ),
            self.s3,
            self.client,
        )

    def get_model_markups(self, pool_id: str) -> List[mapping.AssignmentSolutions]:
        if not self.model_ws:
            return []

        # TODO(DATAFORGE-75): model can work only on first attempt

        assignment = self.model_ws.get_solutions(self.pool_input_objects)

        solution_id_to_evaluation = self.get_checks(pool_id)

        assignment_solutions = mapping.get_assignments_solutions(
            assignments=[assignment],
            mapping=self.markup_task_mapping,
        )

        if solution_id_to_evaluation:
            # needed to set actual assignment status
            evaluation.evaluate_submitted_assignments_and_apply_rules(
                assignment_solutions,
                evaluation.CheckAssignmentAccuracyEvaluationStrategy(
                    solution_id_to_evaluation,
                    self.check_task_mapping,
                ),
                self.markup_params.control,
                self.client,
                self.lang,
                pool_id,
            )

        return assignment_solutions

    def check_markups(
        self,
        markup_pool_id: str,
        check_pool_id: str,
        markup_assignments: List[mapping.AssignmentSolutions],
        fast_markup_assignments: List[mapping.AssignmentSolutions],
    ) -> Optional[CheckMarkupIterationStats]:
        submitted_markup_assignments = [
            (assignment, solutions)
            for assignment, solutions in markup_assignments
            if assignment.status == toloka.Assignment.SUBMITTED
        ]

        if len(submitted_markup_assignments) == 0:
            return None

        logger.debug(f'checking {len(submitted_markup_assignments)} submitted markup assignments')

        solution_id_to_evaluation = self.get_checks(check_pool_id)

        self.create_new_check_tasks(submitted_markup_assignments, solution_id_to_evaluation, check_pool_id)

        solution_id_to_evaluation = self.get_checks(check_pool_id)

        evaluation.evaluate_submitted_assignments_and_apply_rules(
            submitted_markup_assignments,
            evaluation.CheckAssignmentAccuracyEvaluationStrategy(solution_id_to_evaluation, self.check_task_mapping),
            self.markup_params.control,
            self.client,
            self.lang,
            markup_pool_id,
        )

        finalized_tasks = evaluation.find_finalized_tasks(
            markup_assignments,
            solution_id_to_evaluation,
            self.markup_task_mapping,
            self.check_task_mapping,
            self.evaluation.assignment_check_sample,
            self.markup_params.overlap.max_overlap,
        )

        task_id_to_overlap_increase = {
            task_id: 1
            for task_id in (
                {self.markup_task_mapping.task_id(objects) for objects in self.pool_input_objects} - finalized_tasks
            )
        }

        tasks_to_re_markup = classification_loop.rework_not_finalized_tasks(
            client=self.client,
            assignments=markup_assignments + fast_markup_assignments,
            task_mapping=self.markup_task_mapping,
            task_id_to_overlap_increase=task_id_to_overlap_increase,
            pool_id=markup_pool_id,
        )

        ok_tasks_ratio = evaluation.get_ok_tasks_ratio(solution_id_to_evaluation, self.markup_task_mapping)

        return CheckMarkupIterationStats(tasks_to_re_markup, ok_tasks_ratio)

    def get_checks_and_weights(
        self,
        pool_id: str,
    ) -> Tuple[Dict[mapping.TaskID, evaluation.SolutionEvaluation], Optional[Dict[str, float]]]:
        self.check_loop.loop(pool_id)
        _, accepted_assignments, worker_weights = self.check_loop.get_assignments_and_worker_weights(pool_id)
        solution_id_to_evaluation = evaluation.collect_evaluations_from_check_assignments(
            assignments=accepted_assignments,
            check_task_mapping=self.check_task_mapping,
            pool_input_objects=None,
            aggregation_algorithm=self.evaluation.aggregation_algorithm,
            confidence_threshold=self.markup_params.overlap.confidence,
            worker_weights=worker_weights,
        )
        for solution_id, solution_evaluation in solution_id_to_evaluation.items():
            # solutions from previous iterations are also displayed here, may be fix it
            logger.debug(
                f'solution "{solution_id}" is {"OK" if solution_evaluation.ok else "BAD"}, '
                f'confidence={solution_evaluation.confidence:.3f}'
            )
        return solution_id_to_evaluation, worker_weights

    def get_checks(self, pool_id: str) -> Dict[mapping.TaskID, evaluation.SolutionEvaluation]:
        return self.get_checks_and_weights(pool_id)[0]

    def create_new_check_tasks(
        self,
        submitted_markup_assignments: List[mapping.AssignmentSolutions],
        solution_id_to_evaluation: Dict[mapping.TaskID, evaluation.SolutionEvaluation],
        check_pool_id: str,
    ):
        solutions_to_check, already_checked_solutions = evaluation.find_markup_solutions_to_check(
            markup_assignments=submitted_markup_assignments,
            checked_solutions=set(solution_id_to_evaluation.keys()),
            check_task_mapping=self.check_task_mapping,
            check_sample=self.evaluation.assignment_check_sample,
        )

        # TODO: build metric based on already_checked_solutions ratio
        for solution_id in already_checked_solutions:
            logger.debug(f'solution "{solution_id}" is already checked')

        if len(solutions_to_check) == 0:
            return

        self.check_loop.add_input_objects(check_pool_id, solutions_to_check)



import abc
from dataclasses import dataclass, field
import datetime
from dateutil.relativedelta import relativedelta
from typing import List, Optional, Set, Tuple

import toloka.client as toloka

from .common import Worker
from ..utils import and_, or_


# assignment provides us moment in time and other info
@dataclass
class Human(Worker):
    user_id: str
    assignment_id: str

    def __init__(self, assignment: toloka.Assignment):
        self.user_id = assignment.user_id
        self.assignment_id = assignment.id

    @property
    def id(self) -> str:
        return self.user_id


class RegionCodes:
    UNITED_STATES = 84
    GERMANY = 96
    UNITED_KINGDOM = 102
    NETHERLANDS = 118
    FINLAND = 123
    FRANCE = 124
    SWEDEN = 127
    BELARUS = 149
    KAZAKHSTAN = 159
    AZERBAIJAN = 167
    ISRAEL = 181
    UKRAINE = 187
    SPAIN = 204
    ITALY = 205
    RUSSIA = 225
    TURKEY = 983


lang_to_default_regions = {
    'AZ': {RegionCodes.AZERBAIJAN},
    'DE': {RegionCodes.GERMANY},
    'EN': {RegionCodes.UNITED_STATES, RegionCodes.UNITED_KINGDOM},
    'ES': {RegionCodes.SPAIN},
    'FI': {RegionCodes.FINLAND},
    'FR': {RegionCodes.FRANCE},
    'HE': {RegionCodes.ISRAEL},
    'IT': {RegionCodes.ITALY},
    'KK': {RegionCodes.KAZAKHSTAN},
    'NL': {RegionCodes.NETHERLANDS},
    'RU': {RegionCodes.RUSSIA, RegionCodes.BELARUS, RegionCodes.UKRAINE},
    'SV': {RegionCodes.SWEDEN},
    'TR': {RegionCodes.TURKEY},
}


@dataclass
class LanguageRequirement:
    lang: str
    verified: bool = False


@dataclass
class HumanFilter:
    @abc.abstractmethod
    def to_toloka_filter(self) -> Optional[toloka.filter.FilterCondition]:
        ...


@dataclass
class ExpertFilter(HumanFilter):
    skills: List[toloka.Skill]

    def to_toloka_filter(self) -> toloka.filter.FilterCondition:
        return toloka.filter.FilterOr([(toloka.filter.Skill(skill.id) > 0) for skill in self.skills])


@dataclass
class BaseWorkerFilter(HumanFilter, abc.ABC):
    training_score: Optional[int]

    def __post_init__(self):
        assert self.training_score is None or (
            isinstance(self.training_score, int) and (0 <= self.training_score <= 100)
        )


# Actually Toloka allows to make arbitrary filters with AND/OR expressions. Currently, in our code we work with
# specific filters with predefined AND/OR structure, but during lzy serialization we store filters as unstructured
# Toloka object, so we have possibility to make more complex filters in future without loss of backward compatibility.
@dataclass
class WorkerFilter(BaseWorkerFilter):
    @dataclass
    class Params(HumanFilter):
        lang: Optional[LanguageRequirement]
        regions: Set[int] = field(default_factory=set)  # empty means all regions
        age_range: Tuple[Optional[int], Optional[int]] = (18, None)
        client_types: Set[toloka.filter.ClientType.ClientType] = field(default_factory=set)  # empty means all types

        def to_toloka_filter(self) -> Optional[toloka.filter.FilterCondition]:
            filters = []
            if self.lang:
                kwargs = {}
                if self.lang.verified:
                    kwargs['verified'] = True  # we cannot pass verified=False parameter to toloka-kit Languages
                filters.append(toloka.filter.Languages.in_(self.lang.lang, **kwargs))
            if self.regions:
                regions = sorted(self.regions)  # sorted for determinism
                filters.append(or_([toloka.filter.RegionByPhone.in_(code) for code in regions]))
            today = datetime.datetime.today()
            min_age, max_age = self.age_range
            if min_age:
                filters.append(toloka.filter.DateOfBirth < int((today - relativedelta(years=min_age)).timestamp()))
            if max_age:
                filters.append(toloka.filter.DateOfBirth > int((today - relativedelta(years=max_age)).timestamp()))
            if self.client_types:
                client_types = sorted(list(self.client_types), key=lambda i: i.value)  # sorted for determinism
                filters.append(or_([toloka.filter.ClientType == t for t in client_types]))
            return and_(filters)

    filters: List[Params]

    def to_toloka_filter(self) -> Optional[toloka.filter.FilterCondition]:
        return or_([f.to_toloka_filter() for f in self.filters])


@dataclass
class CustomWorkerFilter(BaseWorkerFilter):
    filter: toloka.filter.FilterCondition

    def to_toloka_filter(self) -> toloka.filter.FilterCondition:
        return self.filter



from .. import base, evaluation

acceptance = base.LocalizedString(
    {
        'RU': f"""
<div>
    <input id="hd-1" class="hide" type="checkbox">
    <label for="hd-1"><strong>Нажмите, чтобы изучить информацию о проверке заданий</strong></label>
    <div>
        <p>На проверку отправляются несколько случайно выбранных заданий со страницы. Если доля неверно выполненных
            заданий больше максимальной допустимой, то страница заданий отклоняется, а в сообщение пишутся номера
            неверно выполненных заданий. Также возможен вариант, когда будет оцениваться страница целиком, и она
            будет отклонена, если общее качество страницы ниже допустимого.</p>

        <p>Задания пронумерованы в порядке следования на странице заданий, начиная с 1.</p>

        <div>
            <input id="hd-2" class="hide" type="checkbox">
            <label for="hd-2"><strong>Пример номера задания</strong></label>
            <div>
                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/task-number.jpg"
                     alt="Task number"/>
            </div>
        </div>

        <p>Если вы не согласны с отклонением страницы заданий, подайте апелляцию, написав сообщение в следующем
            формате:</p>

        <ul>
            <li>В первой строке – слово "Апелляция"</li>
            <li>Во второй строке – номер страницы заданий</li>
            <li>В третьей строке – номера неверно выполненных заданий, которые вы считаете выполненными верно</li>
            <li>Начиная с четвертой строки вы можете дать комментарий относительно конкретных заданий или страницы
                заданий в целом
            </li>
        </ul>

        <div>
            <input id="hd-3" class="hide" type="checkbox">
            <label for="hd-3"><strong>Пример при работе из браузера</strong></label>
            <div>
                <p>Проверив раздел "Профиль > История", вы увидите подобную запись:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/ru/rejection-message_browser.png"
                     alt="Rejection message"/>

                <p>Зеленым обведен номер <strong>страницы заданий</strong>, синим – номера неверно выполненных
                    <strong>заданий</strong>.</p>

                <p>Допустим, вы считаете, что вы верно решили задания 9 и 15. Тогда отправьте сообщение в следующем
                    формате:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/ru/appeal-message_browser.png"
                     alt="Appeal message"/>

                <p>Если при написании сообщения есть возможность указать тему, напишите "Апелляция".</p>
            </div>
        </div>
        <div>
            <input id="hd-4" class="hide" type="checkbox">
            <label for="hd-4"><strong>Пример при работе из мобильного приложения</strong></label>
            <div>
                <p>Проверив раздел "Мои задания", вы увидите в списке карточку вида:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/ru/completed-tasks_app.png"
                         alt="Completed tasks"/>

                <p>Нажмите на карточку. На экране задания синим обведены номера неверно выполненных заданий.</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/ru/rejection-message_app.png"
                         alt="Rejection message"/>

                <p>Допустим, вы считаете, что вы верно решили задания 9 и 15. Скопируйте номер страницы заданий, нажав на
                   "Скопировать ID" (обведено зеленым), нажмите на "Написать заказчику" (обведено фиолетовым). Отправьте
                   сообщение в следующем формате:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/ru/appeal-message_app.png"
                     alt="Appeal message"/>
            </div>
        </div>
        <p>В случае, если проводилась проверка страницы в целом, вы увидите сообщение
            "{evaluation.assignment_short_rejection_comment['RU']}". Если вы не согласны с отклонением страницы заданий,
            подайте апелляцию, написав сообщение в описанном выше формате. Указывайте номера тех заданий, которые вы
            считаете выполненными верно.</p>
    </div>
</div>
""",
        'EN': f"""
<div>
    <input id="hd-1" class="hide" type="checkbox">
    <label for="hd-1"><strong>Click to learn the information about solution verification</strong></label>
    <div>
        <p>Several randomly selected tasks from the page are sent for verification. If the percentage of incorrectly
            performed tasks is more than the maximum allowed, then the task page is rejected, and numbers of
            incorrectly performed tasks are written in the message. Alternatively, your assignment will be evaluated
            as a whole and rejected, if general accuracy is below allowed.</p>

        <p>The tasks are numbered in order on the tasks page, starting from 1.</p>

        <div>
            <input id="hd-2" class="hide" type="checkbox">
            <label for="hd-2"><strong>Task number example</strong></label>
            <div>
                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/task-number.jpg"
                     alt="Task number"/>
            </div>
        </div>

        <p>If you do not agree with the rejection of the task page, file an appeal by writing a message in the following
            format:</p>

        <ul>
            <li>In the first line – the word "Appeal"</li>
            <li>In the second line – the number of the task page</li>
            <li>In the third line – the numbers of incorrectly performed tasks that you consider to have been performed
                correctly</li>
            <li>Starting from the fourth line, you can comment on specific tasks or the task page as a whole.</li>
        </ul>

        <div>
            <input id="hd-3" class="hide" type="checkbox">
            <label for="hd-3"><strong>Example when working from the browser</strong></label>
            <div>
                <p>After checking the "Profile > History" section, you will see a similar entry:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/en/rejection-message_browser.png"
                     alt="Rejection message"/>

                <p>Green is the number of the <strong>task page</strong>, blue is the number of incorrectly performed
                    <strong>tasks</strong>.</p>

                <p>Let's say you think that you have correctly performed tasks 6 and 9. Then send a message in the
                    following format:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/en/appeal-message_browser.png"
                     alt="Appeal message"/>

                <p>If it is possible to specify a topic when writing a message, write "Appeal".</p>
            </div>
        </div>
        <div>
            <input id="hd-4" class="hide" type="checkbox">
            <label for="hd-4"><strong>Example when working from a mobile app</strong></label>
            <div>
                <p>After checking the "My Tasks" section, you will see a card like this in the list:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/en/completed-tasks_app.png"
                         alt="Completed tasks"/>

                <p>Click on the card. On the task screen, the numbers of incorrectly performed tasks are outlined in blue.</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/en/rejection-message_app.png"
                         alt="Rejection message"/>

                <p>Let's say you think that you have correctly performed tasks 5 and 9. Copy the task page number by
                    clicking on "Copy ID" (outlined in green), click on "Write to the customer" (outlined in purple).
                    Send the message is in the following format:</p>

                <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/en/appeal-message_app.png"
                     alt="Appeal message"/>
            </div>
        </div>
        <p>If your assignment was evaluated as a whole, you will see a message
            "{evaluation.assignment_short_rejection_comment['EN']}". If you do not agree with the page rejection,
            file an appeal by writing a message in the format described above. Specify the numbers of tasks you
            think you have done correctly.</p>
    </div>
</div>
""",
    }
)



import json
from random import choice
from typing import Optional, Tuple, List, Union

from .. import classification_loop, experts, feedback_loop, mapping, pool as pool_config

import toloka.client as toloka


def get_pool_id(name: Optional[str]) -> Optional[str]:
    if name is None:
        return None
    try:
        with open(f'{name}.json', 'r') as file:
            return json.load(file)['pool_id']
    except (TypeError, FileNotFoundError, KeyError, ValueError):
        return None


def get_classification_pool(
    client: toloka.TolokaClient,
    name: Optional[str],
    loop: Union[experts.ExpertPipeline, classification_loop.ClassificationLoop],
    pool_cfg: Union[pool_config.ExpertConfig, pool_config.ClassificationConfig],
    control_objects: Optional[List[mapping.TaskSingleSolution]] = None,
) -> toloka.Pool:
    pool_id = get_pool_id(name)
    if pool_id is not None:
        return client.get_pool(pool_id)

    if isinstance(loop, experts.ExpertPipeline):
        assert control_objects is None
        pool = loop.create_pool(pool_cfg)
    else:
        pool = loop.create_pool(control_objects, pool_cfg)

    if name is not None:
        with open(f'{name}.json', 'w') as file:
            json.dump({'pool_id': pool.id}, file)

    return pool


def get_annotation_pools(
    client: toloka.TolokaClient,
    name: Optional[str],
    fb_loop: feedback_loop.FeedbackLoop,
    check_pool_cfg: pool_config.ClassificationConfig,
    markup_pool_cfg: pool_config.MarkupConfig,
    control_objects: List[mapping.TaskSingleSolution],
) -> Tuple[toloka.Pool, toloka.Pool]:
    pool_ids = get_pool_ids(name)

    if pool_ids is None:
        markup_pool, check_pool = fb_loop.create_pools(control_objects, markup_pool_cfg, check_pool_cfg)
    else:
        check_pool_id, markup_pool_id = pool_ids
        check_pool = client.get_pool(check_pool_id)
        markup_pool = client.get_pool(markup_pool_id)

    if name is not None:
        with open(f'{name}.json', 'w') as file:
            json.dump({'check_pool_id': check_pool.id, 'markup_pool_id': markup_pool.id}, file)

    return markup_pool, check_pool


def get_pool_ids(name: Optional[str]) -> Optional[Tuple[str, str]]:
    if name is None:
        return None
    try:
        with open(f'{name}.json', 'r') as file:
            data = json.load(file)
            return data['check_pool_id'], data['check_pool_id']
    except (TypeError, FileNotFoundError, KeyError, ValueError):
        return None


def get_swaps(name: Optional[str], input_objects: List[mapping.Objects]) -> List[bool]:
    if name is not None:
        try:
            with open(f'{name}-swaps.json', 'r') as file:
                return json.load(file)
        except (TypeError, FileNotFoundError, KeyError, ValueError):
            pass
    swaps = [choice((True, False)) for _ in input_objects]

    if name is not None:
        with open(f'{name}-swaps.json', 'w') as file:
            json.dump(swaps, file)

    return swaps



from collections import UserDict
from dataclasses import dataclass
from functools import partial
from typing import List, Union, Any, Tuple, Type, Callable, Optional

import toloka.client.project.template_builder as tb

from ..base import (
    Class,
    ClassMeta,
    Label,
    LocalizedString,
    Object,
    BinaryEvaluation,
    EvaluationMeta,
    Metadata,
    ImageAnnotation,
    CLASS_OBJ_FIELD,
    CLASS_TASK_FIELD,
    EVALUATION_TASK_FIELD,
    TextFormat,
    Title,
    ConditionContext,
    AvailableLabels,
    SingleLabel,
    LabelsDisplayType,
    create_input_label_as_text_if,
)
from ..objects import Audio, Text, Image, Video, TextValidation


def gen_hotkey_plugin(key: Any, where: tb.BaseComponent, what: Any) -> tb.HotkeysPluginV1:
    return tb.plugins.HotkeysPluginV1(**{f'key_{key}': tb.SetActionV1(data=where, payload=what)})


def gen_radio_group_hotkeys(
    cls: Type[Label],
    output_path: str = CLASS_TASK_FIELD,
    start_key: int = 0,
) -> List[tb.plugins.HotkeysPluginV1]:
    return [
        gen_hotkey_plugin(i + start_key, tb.OutputData(path=output_path), what=value)
        for i, value in enumerate(cls.possible_values())
        if i + start_key < 10  # only 9 digits for hotkeys
    ]


def gen_text_view(
    content: Union[str, tb.InputData, tb.OutputData],
    text_format: TextFormat,
) -> Union[tb.TextViewV1, tb.MarkdownViewV1]:
    if text_format == TextFormat.MARKDOWN:
        return tb.MarkdownViewV1(content=content)
    else:
        return tb.TextViewV1(content=content)


def gen_text_output_view(
    data: tb.OutputData,
    validation: Optional[TextValidation],
    placeholder: Optional[LocalizedString],
    lang: str,
) -> tb.TextareaFieldV1:
    tb_validation = None
    tb_placeholder = None

    if placeholder:
        tb_placeholder = placeholder[lang]

    if validation:
        tb_validation = tb.SchemaConditionV1(
            schema={
                'type': 'string',
                'pattern': validation.regex[lang],
            },
            hint=validation.hint[lang],
        )
    return tb.TextareaFieldV1(data=data, placeholder=tb_placeholder, validation=tb_validation)


def gen_label_view(
    *args: ...,
    obj_meta: Union[ClassMeta, EvaluationMeta],
    lang: str,
    condition_context: ConditionContext,
    is_output: bool = False,
    **kwargs: ...,
) -> Union[tb.RadioGroupFieldV1, tb.SelectFieldV1, tb.ListViewV1, tb.IfHelperV1]:
    values_labels = list(zip(obj_meta.type.possible_values(), obj_meta.type.get_display_labels()))

    if not isinstance(obj_meta, ClassMeta) or not obj_meta.available_labels:
        return _gen_label_view(
            *args,
            obj_meta=obj_meta,
            values_labels=values_labels,
            lang=lang,
            condition_context=condition_context,
            is_output=is_output,
            **kwargs,
        )

    def labels_to_view(labels: AvailableLabels) -> Union[tb.RadioGroupFieldV1, tb.SelectFieldV1, tb.ListViewV1]:
        filtered_values = [label.value for label in labels.labels]
        filtered = [(value, display_label) for value, display_label in values_labels if value in filtered_values]
        return _gen_label_view(
            *args,
            obj_meta=obj_meta,
            values_labels=filtered,
            lang=lang,
            condition_context=condition_context,
            is_output=is_output,
            **kwargs,
        )

    return obj_meta.available_labels.to_toloka_if(labels_to_view, condition_context)


def _gen_label_view(
    *args: ...,
    obj_meta: Union[ClassMeta, EvaluationMeta],
    values_labels: List[Tuple[Any, LocalizedString]],
    lang: str,
    condition_context: ConditionContext,
    is_output: bool = False,
    **kwargs: ...,
) -> Union[tb.RadioGroupFieldV1, tb.SelectFieldV1, tb.ListViewV1]:
    radio_group_view = tb.RadioGroupFieldV1(
        *args,
        options=[tb.GroupFieldOption(label=label[lang], value=value) for value, label in values_labels],
        disabled=not is_output,
        **kwargs,
    )
    select_field_view = tb.SelectFieldV1(
        *args,
        options=[tb.SelectFieldV1.Option(label=label[lang], value=value) for value, label in values_labels],
        **kwargs,
    )

    many_labels = len(values_labels) > 5

    if is_output:
        if many_labels:
            options_view = select_field_view
        else:
            options_view = radio_group_view
    else:
        as_text = isinstance(obj_meta, ClassMeta) and obj_meta.input_display_type == LabelsDisplayType.MONO
        if as_text or many_labels:

            def text_label_view(label: SingleLabel) -> tb.TextViewV1:
                assert isinstance(label.label, Class)
                return gen_text_view(content=label.label.get_label()[lang], text_format=obj_meta.text_format)

            options_view = create_input_label_as_text_if(
                obj_meta.name, obj_meta.type.possible_instances()
            ).to_toloka_if(
                view_gen=text_label_view,
                condition_context=condition_context,
            )
        else:
            options_view = radio_group_view

    return add_title(options_view, obj_meta.title, lang)


def add_title(view: Any, title: Title, lang: str) -> Union[Any, tb.ListViewV1]:
    if title:
        return tb.GroupViewV1(
            content=tb.ListViewV1(items=[gen_text_view(content=title.text[lang], text_format=title.format), view])
        )
    return view


@dataclass(frozen=True)
class ObjectInfo:
    obj_field: str
    task_field: str
    view_field: str
    input_viewer: Union[Type[tb.view.BaseViewV1], Callable[..., tb.fields.BaseFieldV1]]
    output_viewer: Union[Type[tb.fields.BaseFieldV1], Callable[..., tb.fields.BaseFieldV1]]
    input_validation: Optional[Callable[..., tb.conditions.BaseConditionV1]] = None
    output_validation: Optional[Callable[..., tb.conditions.BaseConditionV1]] = None
    hotkeys: Callable[..., List[tb.plugins.HotkeysPluginV1]] = lambda *args, **kwargs: []


class ViewInfo(UserDict):
    def __getitem__(self, item: Type[Object]) -> ObjectInfo:
        if isinstance(item, type) and issubclass(item, Class):
            return self.data[Class]
        return self.data[item]


def option_choice_required(data: tb.data.BaseData, lang: str) -> tb.RequiredConditionV1:
    return tb.RequiredConditionV1(
        data=data,
        hint=LocalizedString(
            {
                'EN': 'Choose one of the options',
                'RU': 'Выберите один из вариантов',
                'TR': 'Seçeneklerden birini seçin',
                'KK': 'Опциялардың бірін таңдаңыз',
            }
        )[lang],
    )


def image_annotation_required(data: tb.data.BaseData, lang: str) -> tb.RequiredConditionV1:
    return tb.RequiredConditionV1(
        data=data,
        hint=LocalizedString(
            {
                'EN': 'Select at least one area',
                'RU': 'Выделите хотя бы одну область',
                # todo: TR, KK
            }
        )[lang],
    )


def raise_assert_on_output_metadata(*args, **kwargs):
    assert False, 'Metadata objects can only be used as inputs'


object_to_info = ViewInfo(
    {  # Dict[Type[Object], ObjectInfo]
        Audio: ObjectInfo(
            obj_field='url',
            task_field='audio',
            view_field='url',
            input_viewer=tb.AudioViewV1,
            input_validation=lambda *args, **kwargs: tb.PlayedFullyConditionV1(),
            output_viewer=tb.AudioFieldV1,
        ),
        Text: ObjectInfo(
            obj_field='text',
            task_field='text',
            view_field='content',
            input_viewer=gen_text_view,
            output_viewer=gen_text_output_view,
        ),  # TextField may be more suitable in some cases
        Image: ObjectInfo(
            obj_field='url',
            task_field='image',
            view_field='url',
            input_viewer=tb.ImageViewV1,
            output_viewer=lambda *args, **kwargs: tb.MediaFileFieldV1(
                *args, accept=tb.fields.MediaFileFieldV1.Accept(photo=True), **kwargs
            ),
        ),
        Video: ObjectInfo(
            obj_field='url',
            task_field='video',
            view_field='url',
            input_viewer=tb.VideoViewV1,
            output_viewer=lambda *args, **kwargs: tb.MediaFileFieldV1(
                *args, accept=tb.fields.MediaFileFieldV1.Accept(video=True), **kwargs
            ),
        ),
        BinaryEvaluation: ObjectInfo(
            obj_field='ok',
            task_field=EVALUATION_TASK_FIELD,
            view_field='ok',
            input_viewer=gen_label_view,
            output_viewer=partial(gen_label_view, is_output=True),
            output_validation=option_choice_required,
            hotkeys=gen_radio_group_hotkeys,
        ),
        Class: ObjectInfo(
            obj_field=CLASS_OBJ_FIELD,
            task_field=CLASS_TASK_FIELD,
            view_field=CLASS_OBJ_FIELD,
            input_viewer=gen_label_view,
            output_viewer=partial(gen_label_view, is_output=True),
            output_validation=option_choice_required,
            hotkeys=gen_radio_group_hotkeys,
        ),
        Metadata: ObjectInfo(
            obj_field='metadata',
            task_field='metadata',
            view_field='metadata',
            input_viewer=lambda *args, **kwargs: tb.ListViewV1([]),
            output_viewer=raise_assert_on_output_metadata,
        ),
        ImageAnnotation: ObjectInfo(
            obj_field='data',
            task_field='image_annotation',
            view_field='data',
            input_viewer=lambda *args, **kwargs: tb.ImageAnnotationFieldV1(*args, disabled=True, **kwargs),
            output_viewer=tb.ImageAnnotationFieldV1,
            output_validation=image_annotation_required,  # todo: questionable
        ),
    }
)

missing_media_notification_1 = LocalizedString(
    {
        'EN': 'Before performing a task, make sure that all media elements have loaded.',
        'RU': 'Перед выполнением проверьте, что все медиа-элементы загрузились.',
        'TR': 'Çalıştırmadan önce tüm medya öğelerinin yüklendiğini kontrol edin.',
        'KK': 'Іске қоспас бұрын барлық медиа элементтердің жүктелгенін тексеріңіз.',
    }
)
missing_media_notification_2 = LocalizedString(
    {
        'EN': 'If at least one media element is missing, reload the page.',
        'RU': 'Если нет хотя бы одного медиа-элемента, перезагрузите страницу.',
        'TR': 'En az bir medya öğesi yoksa, sayfayı yeniden yükleyin.',
        'KK': 'Егер кем дегенде бір медиа элемент болмаса, бетті қайта жүктеңіз.',
    }
)
media_notifications = [missing_media_notification_1, missing_media_notification_2]

CHECK = LocalizedString({'EN': 'check', 'RU': 'проверка', 'TR': 'kontrol', 'KK': 'тексеру'})



import enum

from .. import base


class ExpertCase(enum.Enum):
    TASK_VERIFICATION = 'task-verification'
    LABELING_QUALITY_VERIFICATION = 'labeling-quality-verification'

    @property
    def label(self) -> base.LocalizedString:
        return base.LocalizedString(
            {
                self.TASK_VERIFICATION: {
                    'EN': 'Task verification',
                    'RU': 'Проверка заданий',
                },
                self.LABELING_QUALITY_VERIFICATION: {
                    'EN': 'Labeling quality verification',
                    'RU': 'Проверка качества разметки',
                },
            }[self]
        )



from datetime import timedelta
import logging
from typing import List

from .. import mapping, pricing

logger = logging.getLogger(__name__)


# some libs can do it, but we don't want to add dependency just because of this function
def ask(question: str) -> bool:
    answer = input(f'{question} [Y/n] ')
    return answer.lower() == 'y'


def ask_or_warn(msg: str, interactive: bool) -> bool:
    if interactive:
        return ask(msg)
    logger.warn(msg)
    return True


OBJECTS_COUNT_VALID_RANGE = (1, 100000)
OBJECTS_COUNT_RECOMMENDED_RANGE = (1, 10000)
OBJECTS_DURATION_RECOMMENDED_RANGE = (timedelta(minutes=1), timedelta(days=2))
CONTROL_OBJECTS_RECOMMENDED_MIN_RATIO = 0.05
RECOMMENDED_RANGE_MESSAGE = (
    'Too small objects volume is inefficient in terms of speed, too big may cause problems '
    'in crowd-sourcing platform API or can be risky if launch is misconfigured'
)


def validate_objects_volume(
    input_objects: List[mapping.Objects],
    control_objects: List[mapping.TaskSingleSolution],  # assisted by us
    task_duration_hint: timedelta,
    pricing_config: pricing.PoolPricingConfig,
    interactive: bool = False,
) -> bool:
    no_control_objects = pricing_config.control_tasks_count == 0

    if not control_objects:
        assert no_control_objects, 'You should provide control objects'

    control_objects_valid_range = (
        max(OBJECTS_COUNT_VALID_RANGE[0], pricing_config.control_tasks_count),
        OBJECTS_COUNT_VALID_RANGE[1],
    )
    control_objects_recommended_range = (
        max(OBJECTS_COUNT_RECOMMENDED_RANGE[0], pricing_config.control_tasks_count),
        OBJECTS_COUNT_RECOMMENDED_RANGE[1],
    )
    control_objects_recommended_duration_range = OBJECTS_DURATION_RECOMMENDED_RANGE
    control_objects_recommended_min_ratio = CONTROL_OBJECTS_RECOMMENDED_MIN_RATIO

    if no_control_objects:
        control_objects_valid_range = control_objects_recommended_range = (0, 0)
        control_objects_recommended_duration_range = (timedelta(0), timedelta(0))
        control_objects_recommended_min_ratio = 0.0

    question = 'Do you wish to continue?'

    for objects, desc, valid_range, recommended_range, duration_recommended_range in (
        (
            input_objects,
            'Objects',
            OBJECTS_COUNT_VALID_RANGE,
            OBJECTS_COUNT_RECOMMENDED_RANGE,
            OBJECTS_DURATION_RECOMMENDED_RANGE,
        ),
        (
            control_objects,
            'Control objects',
            control_objects_valid_range,
            control_objects_recommended_range,
            control_objects_recommended_duration_range,
        ),
    ):
        assert valid_range[0] <= len(objects) <= valid_range[1], (
            f'{desc} count must be in range [{valid_range[0]}, {valid_range[1]}], ' f'you have {len(objects)}'
        )

        if not (recommended_range[0] <= len(objects) <= recommended_range[1]):
            msg = (
                f'{desc} count recommended range is [{recommended_range[0]}, {recommended_range[1]}], '
                f'you have {len(objects)}. {RECOMMENDED_RANGE_MESSAGE}. {question}'
            )
            if not ask_or_warn(msg, interactive):
                return False

        if not (duration_recommended_range[0] <= task_duration_hint * len(objects) <= duration_recommended_range[1]):
            msg = (
                f'{desc} duration recommended range is '
                f'[{duration_recommended_range[0]} - {duration_recommended_range[1]}], '
                f'you have {task_duration_hint * len(objects)}. {RECOMMENDED_RANGE_MESSAGE}. {question}'
            )
            if not ask_or_warn(msg, interactive):
                return False

    if len(control_objects) / len(input_objects) < control_objects_recommended_min_ratio:
        msg = (
            f'Recommended control objects minimum ratio is {CONTROL_OBJECTS_RECOMMENDED_MIN_RATIO}, you have '
            f'{len(control_objects) / len(input_objects):.2f}. A sufficient control tasks ratio is needed for '
            f'higher throughput (each worker will see specific control task only once). {question}'
        )
        if not ask_or_warn(msg, interactive):
            return False

    return True



from collections import UserDict
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
import logging
from math import floor, ceil
import random
import threading
from typing import List, Dict, Any, Tuple, Optional, Union

import numpy as np
import toloka.client as toloka

from dataforge import base, client, mapping, objects

logger = logging.getLogger(__name__)


class StoppableThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    def run(self):
        delay = np.random.randint(0, 60)
        self._stop_event.wait(timeout=delay)
        while not self.stopped():
            no_tasks = self._target(*self._args, **self._kwargs)
            if no_tasks:
                self._stop_event.wait(timeout=10 + np.random.randint(0, 10))


EPS = 1e-6


# вероятности плохо работают с крайними значениями
def clip(x: float, eps: float = EPS) -> float:
    return np.clip(x, eps, 1 - eps)


def get_task_id(task_spec: client.PreparedTaskSpec, row: Dict[str, Any]) -> str:
    task_mapping = task_spec.task_mapping
    return task_mapping.task_id(task_mapping.from_task_values(row)).id


class WorkerSpeed(Enum):
    RAND = 'rand'
    POOR = 'poor'
    FAST = 'fast'
    MEDIUM = 'medium'
    SLOW = 'slow'

    def get_bounds(self) -> Tuple[float, float]:
        return {
            WorkerSpeed.RAND: (0.01, 0.05),
            WorkerSpeed.POOR: (0.2, 0.25),
            WorkerSpeed.FAST: (0.5, 0.8),
            WorkerSpeed.MEDIUM: (0.9, 1.1),
            WorkerSpeed.SLOW: (1.2, 1.5),
        }[self]

    # todo - тут надо учитывать еще сколько заданий на странице, а не просто hint
    def get_timeout(self, task_duration_hint: timedelta):
        left, right = self.get_bounds()
        duration = int(task_duration_hint.total_seconds())
        return np.random.randint(floor(left * duration), ceil(right * duration))


@dataclass
class TaskInfo:
    solution: base.Object

    # сложность заданий
    complexity: Optional[float] = 0.0


class AnnotationTaskInfo(TaskInfo):
    solution: mapping.Objects


class AnnotationSolutionsDict(UserDict):
    markup_task_mapping: mapping.TaskMapping
    check_task_mapping: mapping.TaskMapping

    # (task, true solution, [given_solution])
    def __init__(self, data, markup_task_mapping: mapping.TaskMapping, check_task_mapping: mapping.TaskMapping):
        super(AnnotationSolutionsDict, self).__init__(data)
        self.markup_task_mapping = markup_task_mapping
        self.check_task_mapping = check_task_mapping

    def __getitem__(
        self,
        task_id: mapping.TaskID,
    ) -> AnnotationTaskInfo:

        if len(task_id.objects) == len(self.check_task_mapping.input_mapping):
            objects = task_id.objects[: len(self.markup_task_mapping.input_mapping)]
            key = mapping.TaskID(objects)
            given_solution = task_id.objects[len(self.markup_task_mapping.input_mapping) :]
        else:
            given_solution = None
            key = task_id
        info = self.data[key]

        return AnnotationTaskInfo(solution=(key.objects, info.solution, given_solution), complexity=info.complexity)


def get_random_classification_option(
    task_mapping: mapping.TaskMapping,
    forbidden_solution: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    assert len(task_mapping.output_mapping) == 1
    label_class = task_mapping.output_mapping[0].obj_meta.type
    assert issubclass(label_class, base.Class)

    options = []

    # todo we can have confusion matrix here
    for possible_result in label_class.possible_instances():
        option = task_mapping.toloka_values((possible_result,), output=True)
        if forbidden_solution is None or option != forbidden_solution:
            options.append(option)

    return np.random.choice(options)


class TaskComplexity(Enum):
    EASY = 0.05
    OK = 0.5
    HARD = 5


def swap_task(task: mapping.Objects, task_spec: client.PreparedTaskSpec) -> mapping.Objects:
    function = task_spec.function
    assert isinstance(function, base.SbSFunction)
    assert len(task_spec.task_mapping.output_mapping) == 1
    hint_count = len(function.hints)
    hints = task[:hint_count]
    input_objects = task[hint_count:]
    left_objects, right_objects = (input_objects[: len(input_objects) // 2], input_objects[len(input_objects) // 2 :])
    return hints + right_objects + left_objects


# todo we are working with classification functions only
def create_tasks_info_dict(
    task_spec: client.PreparedTaskSpec,
    solved_tasks: List[mapping.TaskSingleSolution],
    tasks_complexity: Optional[Union[Dict[TaskComplexity, int], TaskComplexity]] = None,
) -> Dict[mapping.TaskID, TaskInfo]:

    if tasks_complexity is None:
        tasks_complexity = {TaskComplexity.EASY: len(solved_tasks)}
    elif isinstance(tasks_complexity, TaskComplexity):
        tasks_complexity = {tasks_complexity: len(solved_tasks)}
    else:
        assert isinstance(tasks_complexity, dict)

    assert sum(tasks_complexity.values()) == len(solved_tasks)
    complexities = [c for c, v in tasks_complexity.items() for _ in range(v)]
    solutions = {}

    if isinstance(task_spec.function, base.ClassificationFunction) or isinstance(task_spec.function, base.SbSFunction):
        for (task, solution), complexity in zip(solved_tasks, complexities):
            assert len(solution) == 1
            choice = solution[0]

            solutions[mapping.TaskID(task)] = TaskInfo(choice, complexity.value)
            if isinstance(task_spec.function, base.SbSFunction):
                solutions[mapping.TaskID(swap_task(task, task_spec))] = TaskInfo(choice.swap(), complexity.value)
        return solutions

    assert isinstance(task_spec, client.AnnotationTaskSpec)

    for (task, solution), complexity in zip(solved_tasks, complexities):
        solutions[mapping.TaskID(task)] = AnnotationTaskInfo(solution, complexity.value)

    return AnnotationSolutionsDict(solutions, task_spec.task_mapping, task_spec.check.task_mapping)


class ThreadWithReturnValue(threading.Thread):
    def __init__(self, *args, **kwargs):
        super(ThreadWithReturnValue, self).__init__(*args, **kwargs)
        self._return_value: Any = None

    def run(self):
        self._return_value = self._target(*self._args, **self._kwargs)

    def join(self, timeout: Optional[float] = None) -> Any:
        super(ThreadWithReturnValue, self).join(timeout)
        return self._return_value


def get_accuracy(assignment: toloka.Assignment) -> float:
    ok_count, total_count = 0, 0
    for task, solution in zip(assignment.tasks, assignment.solutions):
        if task.known_solutions:
            total_count += 1
            ok_count += int(solution.output_values == task.known_solutions[0].output_values)
    return ok_count / total_count


def content(object: objects.Object) -> str:
    if isinstance(object, objects.Text):
        return object.text
    elif isinstance(object, objects.Audio):
        return object.url
    raise ValueError('Unknown object type')


# just as an example
def mutate(gt: mapping.Objects) -> mapping.Objects:
    result = []
    for obj in gt:
        if isinstance(obj, objects.Text):
            result.append(mutate_text(obj))
        else:
            raise NotImplementedError(f'{gt}')
    return tuple(result)


# todo: maybe reuse Mutator here
def mutate_text(gt: objects.Text) -> objects.Text:
    letters = list(gt.text)
    random.shuffle(letters)
    return objects.Text(''.join(letters))



import abc
from collections import defaultdict
from enum import Enum
from typing import List, Dict, Type, Tuple

from ..base import Class


class Question(Class):
    # user-defined class. contains both the question and all possible answers for it
    # possible answers are stored in enum values, question_label returns the question itself

    @classmethod
    @abc.abstractmethod
    def question_label(cls) -> Dict[str, str]:
        ...


class CombinedAnswer(Class):
    # auto-generated class. uniquely combines all possible answer values from several questions
    question_mapping: Dict['CombinedAnswer', Question]

    def get_original_answer(self) -> Question:
        # one of the answer options from original Question class, can be used is Results
        return self.question_mapping[self]


def get_combined_classes(
    questions: List[Type[Question]],
) -> Tuple[Type[Class], Type[CombinedAnswer], List[Tuple[Class, List[Class]]]]:
    # we need to combine all possible answers into a single CombinedAnswer Class with unique answer variants
    # and combine all possible questions into a single Class

    question_fields = {}  # {field_name -> field_value} for new Question Enum
    question_labels = {}  # {question_name -> question_label}

    answer_fields = {}  # {field_name -> field_value} for new CombinedAnswer Enum
    answer_values = []  # all Enum instances with possible answers from all given Questions

    question_answers_dict = defaultdict(list)
    for i, question in enumerate(questions):
        question_name = question.__name__
        enum_name = question_name.upper()
        enum_value = question_name.lower()
        question_fields[enum_name] = enum_value
        question_labels[enum_name] = question.question_label()
        for answer in list(question):

            new_value = f'{enum_value}__{answer.value}'  # ensuring new values are unique
            new_name = f'{enum_name}__{answer.name}'  # and new names also

            answer_values.append(answer)
            answer_fields[new_name] = new_value
            question_answers_dict[enum_name].append(new_name)

    GeneratedAnswer = Enum('GeneratedAnswer', answer_fields)

    question_mapping = {}
    for answer, new_answer_enum_value in zip(answer_values, list(GeneratedAnswer)):
        question_mapping[new_answer_enum_value] = answer

    setattr(GeneratedAnswer, 'question_mapping', question_mapping)

    # todo: maybe we should check that set of languages is the same for all question classes
    def answer_labels() -> Dict[GeneratedAnswer, Dict[str, str]]:
        return {answer: answer.get_original_answer().get_label().lang_to_text for answer in list(GeneratedAnswer)}

    setattr(GeneratedAnswer, 'labels', answer_labels)

    GeneratedAnswer.__bases__ = (
        CombinedAnswer,
        Class,
    ) + GeneratedAnswer.__bases__

    CombinedQuestion = Enum('CombinedQuestion', question_fields)

    def questions_labels() -> Dict[CombinedQuestion, Dict[str, str]]:
        return {question: question_labels[question.name.upper()] for question in list(CombinedQuestion)}

    setattr(CombinedQuestion, 'labels', questions_labels)

    CombinedQuestion.__bases__ = (Class,) + CombinedQuestion.__bases__

    question_answers = [
        (
            getattr(CombinedQuestion, question_name),
            [getattr(GeneratedAnswer, answer_name) for answer_name in answer_names],
        )
        for question_name, answer_names in question_answers_dict.items()
    ]
    return CombinedQuestion, GeneratedAnswer, question_answers



import collections
import datetime
from collections import defaultdict
from dataclasses import dataclass
import logging
from typing import List, Tuple, Dict, Optional, Union, Set

import numpy as np
import toloka.client as toloka

from .. import base, classification, control, duration, evaluation, mapping
from ..worker import Human

logger = logging.getLogger(__name__)

# MOS implementation from https://www.microsoft.com/en-us/research/wp-content/uploads/2011/05/0002416.pdf

MIN_ASSIGNMENTS: int = 1  # todo: 10 as default in paper code, maybe should be a dynamic param
MIN_CORRELATION_FOR_ACCEPT: float = 0.25
MIN_CORRELATION_FOR_FINALIZATION: float = 0.7
EPS: float = 1e-10

STUB_CORRELATION: float = MIN_CORRELATION_FOR_ACCEPT + EPS
assert STUB_CORRELATION < MIN_CORRELATION_FOR_FINALIZATION


@dataclass
class ObjectsMetadata:
    item_id: str  # e.g. text that is being synthesised
    algorithm: str  # e.g. speech synthesis model


def get_solutions(
    assignment_solutions: List[mapping.AssignmentSolutions],
) -> List[Tuple[mapping.Objects, classification.WorkerLabel]]:
    tasks = []

    for assignment, solutions in assignment_solutions:
        for solution in solutions:
            (inputs, (score,)) = solution
            assert isinstance(score, base.Label)
            tasks.append((inputs, (score, Human(assignment))))
    return tasks


def t95(n: int, p=0.05) -> float:
    from scipy.stats import t

    return t.ppf(0.5 * (1 + (1.0 - p)), n)


class Stats:
    def __init__(self, values: List[float]):
        self.values = np.array(values)
        self.count = len(values)
        self.mean = np.mean(self.values)
        self.min = np.min(self.values)
        self.max = np.max(self.values)
        self.var = np.var(self.values)  # todo: check var vs std bias
        self.std = np.std(self.values)
        # self.kurtosis = sps.kurtosis(self.values)
        self.ci = None
        if self.count > 1:
            self.ci = t95(self.count - 1) * np.sqrt(self.var / self.count)


class StatDict(collections.UserDict):
    def __init__(self):
        super().__init__({})
        self.stat_values = []
        self.calc_stats = None

    def calculate(self):
        self.calc_stats = Stats(self.stat_values)


class StatList(collections.UserList):
    def __init__(self):
        super().__init__([])
        self.calc_stats = None


def correlation_coefficient(v1: Union[List[float], np.array], v2: Union[List[float], np.array]) -> float:
    from scipy.stats import moment

    assert len(v1) == len(v2)
    if len(v1) <= 1:
        return np.nan
    m1 = np.mean(v1)
    m2 = np.mean(v2)
    count = len(v1)

    sigma1 = np.sqrt(moment(v1, 2))
    sigma2 = np.sqrt(moment(v2, 2))

    sigma12 = 0.0
    for i in range(count):
        sigma12 += (v1[i] - m1) * (v2[i] - m2)
    sigma12 /= count
    return sigma12 / (sigma1 * sigma2 + EPS)


@dataclass
class MOSCI:
    mu: float
    ci: float

    def __post_init__(self):
        self.mu = round(float(self.mu), 2)
        self.ci = round(float(self.ci), 2)

    @property
    def left(self):
        return self.mu - self.ci

    @property
    def right(self):
        return self.mu + self.ci


@dataclass
class AssignmentSet:
    workers: List[str]
    items: List[str]
    algorithms: List[str]
    items_stats: StatDict
    worker_stats: StatDict
    assignments: List[mapping.AssignmentSolutions]

    def worker_correlation(self, worker: str) -> float:
        a_correlation = self.algorithm_mos_correlation(worker)
        s_correlation = self.item_mos_correlation(worker)
        if np.isnan(a_correlation) and np.isnan(s_correlation):
            # edge case: can happen if one worker gets only 1 partial page with one task
            # thus, we only have one submitted answer from them.
            # we should accept the assignment, but not count it towards final score
            return STUB_CORRELATION
        return a_correlation if not np.isnan(a_correlation) else s_correlation

    def item_mos_correlation(self, worker: str) -> float:
        v1, v2 = [], []
        for item in self.items:
            for algorithm in self.algorithms:
                if (
                    algorithm not in self.items_stats
                    or item not in self.items_stats[algorithm]
                    or worker not in self.items_stats[algorithm][item]
                ):
                    continue

                v1.append(self.items_stats[algorithm][item].calc_stats.mean)
                v2.append(self.items_stats[algorithm][item][worker].calc_stats.mean)

        return correlation_coefficient(v1, v2)

    def algorithm_mos_correlation(self, worker: str) -> float:
        v1, v2 = [], []
        for algorithm in self.algorithms:
            if algorithm not in self.worker_stats or worker not in self.worker_stats[algorithm]:
                continue

            v1.append(self.worker_stats[algorithm].calc_stats.mean)
            v2.append(self.worker_stats[algorithm][worker].calc_stats.mean)
        return correlation_coefficient(v1, v2)

    def fast_workers(
        self,
        fast_submits_predicate: control.AssignmentDurationPredicate,
        task_duration_function: duration.TaskDurationFunction,
        task_mapping: mapping.TaskMapping,
    ) -> Set[str]:
        fast_workers = set()
        for assignment, _ in self.assignments:
            assignment_duration = assignment.submitted - assignment.created
            assignment_duration_hint = sum(
                [task_duration_function(task_mapping.from_task(task)) for task in assignment.tasks],
                start=datetime.timedelta(seconds=0),
            )
            if fast_submits_predicate.check(
                assignment_duration,
                assignment_duration_hint=assignment_duration_hint,
            ):
                fast_workers.add(assignment.user_id)

        return fast_workers

    def bad_workers(self) -> Set[str]:
        # todo: there was unused min_assignments param
        bad_workers = set()
        for worker in self.workers:
            if self.worker_correlation(worker) < MIN_CORRELATION_FOR_ACCEPT:
                bad_workers.add(worker)
        return bad_workers

    def outliers(self) -> Set[str]:
        outliers = set()
        worker_to_assignment_count = defaultdict(int)
        for assignment, _ in self.assignments:
            worker_to_assignment_count[assignment.user_id] += 1

        for worker in self.workers:
            if (
                worker_to_assignment_count[worker] < MIN_ASSIGNMENTS
                or self.worker_correlation(worker) < MIN_CORRELATION_FOR_FINALIZATION
            ):
                outliers.add(worker)

        return outliers

    def get_algorithm_stats(self) -> Dict[str, np.array]:
        results = {}
        for algorithm in self.algorithms:
            m = []
            for w in self.workers:
                v = []
                for item in self.items:

                    if (
                        algorithm in self.items_stats
                        and item in self.items_stats[algorithm]
                        and w in self.items_stats[algorithm][item]
                        and self.items_stats[algorithm][item][w].calc_stats.mean
                    ):
                        v.append(self.items_stats[algorithm][item][w].calc_stats.mean)
                    else:
                        v.append(np.nan)
                m.append(v)

            results[algorithm] = np.array(m)

        return results

    def get_algorithm_ci(self) -> Dict[str, MOSCI]:
        algorithm_to_matrix = self.get_algorithm_stats()
        results = {}
        for algorithm in self.algorithms:
            matrix = algorithm_to_matrix[algorithm]
            t = t95(min(matrix.shape))
            mos = np.nanmean(matrix)
            ci95 = t * np.sqrt(mos_variance(matrix))

            results[algorithm] = MOSCI(mos, ci95)

        return results


def mos_variance(Z: np.array) -> np.array:
    N, M = Z.shape

    W = np.logical_not(np.isnan(Z))

    Mi = np.sum(W, 0)
    Nj = np.sum(W, 1)
    T = np.sum(W)

    v_su = vertical_var(Z.T)
    v_wu = vertical_var(Z)
    v_swu = np.nanvar(Z)

    if not np.isnan(v_su) and not np.isnan(v_wu):
        v = np.linalg.solve([[1.0, 0.0, 1.0], [0.0, 1.0, 1.0], [1.0, 1.0, 1.0]], [v_su, v_wu, v_swu])
        v = np.maximum(v, 0)
        v_s, v_w, v_u = v[0], v[1], v[2]
        v_mu = v_s * np.sum(Mi**2) / T**2 + v_w * np.sum(Nj**2) / T**2 + v_u / T
    elif np.isnan(v_su) and not np.isnan(v_wu):
        v = np.linalg.solve([[0.0, 1.0], [1.0, 1.0]], [v_wu, v_swu])
        v = np.maximum(v, 0)

        v_s, v_wu = v[0], v[1]
        v_mu = v_s * np.sum(Mi**2) / T**2 + v_wu / T

    elif not np.isnan(v_su) and np.isnan(v_wu):
        v = np.linalg.solve([[0.0, 1.0], [1.0, 1.0]], [v_su, v_swu])
        v = np.maximum(v, 0)
        v_w, v_su = v[0], v[1]

        v_mu = v_w * np.sum(Nj**2) / T**2 + v_su / T
    else:
        assert np.isnan(v_su) and np.isnan(v_wu)
        v_mu = v_swu / T

    return v_mu


def nancount(vec: np.array) -> int:
    return np.sum(np.logical_not(np.isnan(vec)))


def vertical_var(Z: np.array) -> np.array:
    rows, cols = Z.shape
    v = []
    for i in range(cols):
        if nancount(Z[:, i]) >= 2:
            v.append(np.nanvar(Z[:, i]))
    return np.mean(v)


def _build_stats(
    stats: Dict[str, Dict[str, Dict[str, List[float]]]], k1: List[str], k2: List[str], k3: List[str]
) -> StatDict:
    new_stats = StatDict()
    for key_1 in k1:
        if key_1 not in stats:
            continue
        if key_1 not in new_stats:
            new_stats[key_1] = StatDict()
        for key_2 in k2:
            if key_2 not in stats[key_1]:
                continue
            if key_2 not in new_stats[key_1]:
                new_stats[key_1][key_2] = StatDict()
            for key_3 in k3:
                if key_3 not in stats[key_1][key_2]:
                    continue
                if key_3 not in new_stats[key_1][key_2]:
                    new_stats[key_1][key_2][key_3] = StatList()
                calc_stats = Stats(stats[key_1][key_2][key_3])
                new_stats[key_1][key_2][key_3].calc_stats = calc_stats
                new_stats[key_1][key_2][key_3].extend(stats[key_1][key_2][key_3])
                new_stats[key_1][key_2].stat_values.append(calc_stats.mean)
                new_stats[key_1].stat_values.append(calc_stats.mean)

            new_stats[key_1][key_2].calculate()

        new_stats[key_1].calculate()

    return new_stats


def compute_algorithm_ci95(
    algorithms: List[str], items: List[str], workers: List[str], stats_items: StatDict, stats_workers: StatDict
):
    for algorithm in algorithms:
        if algorithm not in stats_items:
            continue

        t = 0
        mi = []
        nj = []

        v_wu = []

        for item in items:
            if item not in stats_items[algorithm]:
                continue

            if stats_items[algorithm][item].calc_stats.count >= 2:
                v_wu.append(stats_items[algorithm][item].calc_stats.var)
                mi.append(stats_items[algorithm][item].calc_stats.count)
        if v_wu:
            v_wu = np.mean(v_wu)

        v_su = []

        for w in workers:
            if w not in stats_workers[algorithm]:
                continue

            if stats_workers[algorithm][w].calc_stats.count >= 2:
                v_su.append(stats_workers[algorithm][w].calc_stats.var)
                nj.append(stats_workers[algorithm][w].calc_stats.count)

        if v_su:
            v_su = np.mean(v_su)
        v_swu = []

        for item in items:
            if item not in stats_items[algorithm]:
                continue

            for w in workers:
                if w not in stats_items[algorithm][item]:
                    continue
                v_swu.append(stats_items[algorithm][item][w].calc_stats.mean)
                t += 1

        v_swu = np.var(v_swu)
        mi2 = (np.array(mi) ** 2).sum()
        nj2 = (np.array(nj) ** 2).sum()

        if v_su and v_wu:
            m = np.array([[1.0, 0.0, 1.0], [0.0, 1.0, 1.0], [1.0, 1.0, 1.0]])
            c = np.array([[v_su], [v_wu], [v_swu]])
            v = np.linalg.inv(m) @ c
            v_s = max(v[0], 0.0)
            v_w = max(v[1], 0.0)
            v_u = max(v[2], 0.0)
            v_mu = v_s * mi2 / (t**2) + v_w * nj2 / (t**2) + v_u / t
        elif not v_su and v_wu:
            m = np.array([[0.0, 1.0], [1.0, 1.0]])
            c = np.array([[v_wu], [v_swu]])
            v = np.linalg.inv(m) @ c
            v_s = max(v[0], 0.0)
            v_wu = max(v[1], 0.0)
            v_mu = v_s * mi2 / (t**2) + v_wu / t

        elif v_su and not v_wu:
            m = np.array([[0.0, 1.0], [1.0, 1.0]])
            c = np.array([[v_su], [v_swu]])
            v = np.linalg.inv(m) @ c
            v_w = max(v[0], 0.0)
            v_su = max(v[1], 0.0)
            v_mu = v_w * nj2 / (t**2) + v_su / t
        else:
            assert not v_su and not v_wu
            v_mu = v_swu / t

        t = t95(min(len(items), len(workers)) - 1)
        stats_items[algorithm].calc_stats.ci = t * np.sqrt(v_mu)
        stats_workers[algorithm].calc_stats.ci = t * np.sqrt(v_mu)


def build_stats(
    assignments: List[mapping.AssignmentSolutions],
    inputs_to_metadata: Dict[mapping.Objects, ObjectsMetadata],
) -> AssignmentSet:
    tasks = get_solutions(assignments)

    workers = list(set(w.user_id for (_, (_, w)) in tasks))
    items = list(set(inputs_to_metadata[inputs].item_id for (inputs, _) in tasks))
    algorithms = list(set(inputs_to_metadata[inputs].algorithm for (inputs, _) in tasks))
    # stats_workers[algorithm][worker][sentence] = score
    stats_workers = defaultdict(lambda: defaultdict(dict))

    # stats_items[algorithm][sentence][worker] = score
    stats_items = defaultdict(lambda: defaultdict(dict))

    for (inputs, (score, w)) in tasks:
        # task_id = task_mapping.task_id(inputs).id
        assert isinstance(score, base.ScoreEvaluation)
        metadata = inputs_to_metadata[inputs]
        item, algorithm = metadata.item_id, metadata.algorithm
        stats_workers[algorithm][w.user_id][item] = [int(score.value)]
        stats_items[algorithm][item][w.user_id] = [int(score.value)]

    new_stats_items = _build_stats(stats_items, algorithms, items, workers)
    new_stats_workers = _build_stats(stats_workers, algorithms, workers, items)
    compute_algorithm_ci95(algorithms, items, workers, new_stats_items, new_stats_workers)
    return AssignmentSet(workers, items, algorithms, new_stats_items, new_stats_workers, assignments)


class KeyDefaultdict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


@dataclass
class MOSEvaluationStrategy(evaluation.CustomEvaluationStrategy):
    inputs_to_metadata: Optional[Dict[mapping.Objects, ObjectsMetadata]]
    task_mapping: mapping.TaskMapping
    fast_submits_predicate: control.AssignmentDurationPredicate
    task_duration_function: duration.TaskDurationFunction
    updated: bool = False

    def __post_init__(self):
        if self.inputs_to_metadata is None:
            self.inputs_to_metadata = KeyDefaultdict(
                lambda items: ObjectsMetadata(self.task_mapping.task_id(items).id, 'default')
            )

    def ok(self, task: toloka.Task, solution: mapping.TaskSingleSolution) -> Optional[bool]:
        # abc restriction
        return None

    def update(self, submitted_assignments: List[mapping.AssignmentSolutions]):
        assignment_set = build_stats(submitted_assignments, self.inputs_to_metadata)
        fast_workers = assignment_set.fast_workers(
            self.fast_submits_predicate, self.task_duration_function, self.task_mapping
        )
        bad_workers = assignment_set.bad_workers()
        # todo: add logging?
        # todo: maybe reject only bad workers or fast _assignments_
        self.rejected_workers = fast_workers | bad_workers

        filtered_assignments = [
            (assignment, solution)
            for (assignment, solution) in submitted_assignments
            if assignment.user_id not in self.rejected_workers
        ]
        filtered_assignment_set = build_stats(filtered_assignments, self.inputs_to_metadata)

        outliers = filtered_assignment_set.outliers()

        self.final_assignments = [
            (assignment, solution)
            for (assignment, solution) in filtered_assignments
            if assignment.user_id not in outliers
        ]
        self.final_assignment_set = build_stats(self.final_assignments, self.inputs_to_metadata)

        self.updated = True

    def evaluate_assignment(self, assignment_solutions: mapping.AssignmentSolutions) -> evaluation.AssignmentEvaluation:
        assignment, solutions = assignment_solutions
        if not self.updated:
            # can happen inside metrics before end of loop iteration
            return evaluation.AssignmentEvaluation(
                assignment=assignment,
                ok_checks=0,
                total_checks=0,
                objects=len(solutions),
                incorrect_solution_indexes=[],
            )
        ok_checks = 0 if assignment.user_id in self.rejected_workers else 1

        return evaluation.AssignmentEvaluation(
            assignment=assignment,
            ok_checks=ok_checks,
            total_checks=1,
            objects=len(solutions),
            incorrect_solution_indexes=[],
        )



import decimal
import json
from functools import reduce
import logging
import threading
import time
from typing import List, Optional

import toloka.client as toloka

logger = logging.getLogger(__name__)


def get_pool_link(pool: toloka.Pool, client: toloka.TolokaClient) -> str:
    url_prefix = client.url[: -len('/api')]
    return f'{url_prefix}/requester/project/{pool.project_id}/pool/{pool.id}'


# Performs graceful shutdown in case of keyboard interrupt - closes pools if they are already created.
# Shutdown may be called automatically by processing KeyboardInterrupt in case of labeling in main thread,
# or be called manually if for example labeling is running in separate thread.
class LaunchInterrupter:
    client: toloka.TolokaClient
    interrupted: threading.Event
    shut: bool
    pool_ids: List[str]

    def __init__(self, client: toloka.TolokaClient, interrupted: threading.Event):
        self.client = client
        self.interrupted = interrupted
        self.shut = False
        self.pool_ids = []

    def add_pool_id(self, pool_id: str):
        self.pool_ids.append(pool_id)

    def shutdown(self, reraise: bool = True):
        if self.shut:
            return
        for pool_id in self.pool_ids:
            self.client.close_pool(pool_id)
            logger.debug(f'pool {pool_id} is closed')
        logger.info('labeling is canceled')
        self.shut = True
        if reraise:
            raise KeyboardInterrupt


def wait_pool_for_close(
    client: toloka.TolokaClient,
    pool_id: str,
    interrupter: Optional[LaunchInterrupter] = None,
    pull_interval_seconds: float = 60.0,
):
    pool = client.get_pool(pool_id)
    while not pool.is_closed():
        if interrupter and interrupter.interrupted.is_set():
            interrupter.shutdown()  # will raise KeyboardInterrupt to unwind stack up to launch()
        logger.debug(f'waiting pool {get_pool_link(pool, client)} for close...')
        time.sleep(pull_interval_seconds)
        pool = client.get_pool(pool.id)


# 'reduce' is used to avoid redundant AND/OR's with only one element
def and_(filters: List[Optional[toloka.filter.FilterCondition]]) -> Optional[toloka.filter.FilterCondition]:
    filters = [f for f in filters if f is not None]
    if not filters:
        return None
    return reduce(lambda x, y: x & y, filters)


# 'reduce' is used to avoid redundant AND/OR's with only one element
def or_(filters: List[Optional[toloka.filter.FilterCondition]]) -> Optional[toloka.filter.FilterCondition]:
    filters = [f for f in filters if f is not None]
    if not filters:
        return None
    return reduce(lambda x, y: x | y, filters)


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        return super(DecimalEncoder, self).default(o)



from dataclasses import dataclass, field
import io
import logging
from multiprocessing.pool import ThreadPool
import os
from typing import List, Optional, Tuple, Any

import boto3
from botocore.config import Config
import toloka.client as toloka

from .. import base, mapping

logger = logging.getLogger(__name__)


@dataclass
class Attachment:
    id: str
    need_upload: bool
    meta: Optional[toloka.Attachment] = None


@dataclass
class S3:
    endpoint: str
    bucket: str
    path: str

    access_key_id: Optional[str] = None
    secret_access_key: Optional[str] = None

    base_url: str = field(init=False)
    client: Any = field(init=False)

    def __post_init__(self):
        endpoint_url = f'https://{self.endpoint}'
        self.base_url = f'{endpoint_url}/{self.bucket}/{self.path}'
        self.client = boto3.session.Session().client(
            service_name='s3',
            endpoint_url=endpoint_url,
            aws_access_key_id=self.access_key_id or os.getenv('AWS_ACCESS_KEY_ID'),
            aws_secret_access_key=self.secret_access_key or os.getenv('AWS_SECRET_ACCESS_KEY'),
            config=Config(retries=dict(max_attempts=30)),
        )

    def get_url(self, attachment: Attachment) -> str:
        assert attachment.meta
        return f'{self.base_url}/{attachment.meta.name}'

    def upload(self, attachment: Attachment, body: bytes):
        assert attachment.meta
        self.client.put_object(
            Bucket=self.bucket,
            Key=f'{self.path}/{attachment.meta.name}',
            Body=body,
            ACL='public-read',
            ContentType=attachment.meta.media_type,
        )


# replace Toloka attachments with user-configured S3 urls
def substitute_media_output(
    assignments: List[mapping.AssignmentSolutions],
    s3: S3,
    toloka_client: toloka.TolokaClient,
) -> List[mapping.AssignmentSolutions]:
    attachments = _find_or_substitute_media(assignments, None, s3)[1]
    attachments = _get_media_meta(toloka_client, attachments)
    _upload_media(s3, toloka_client, attachments)
    return _find_or_substitute_media(assignments, attachments, s3)[0]


def has_media_output(task_function: base.TaskFunction) -> bool:
    return any(obj_meta.type.is_media() for obj_meta in task_function.get_outputs())


def _find_or_substitute_media(
    assignments: List[mapping.AssignmentSolutions],
    uploaded_attachments: Optional[List[Attachment]],
    s3: S3,
) -> Tuple[List[mapping.AssignmentSolutions], List[Attachment]]:
    new_assignments, found_attachments = [], []
    id_to_uploaded_attachment = (
        {attachment.id: attachment for attachment in uploaded_attachments} if uploaded_attachments else None
    )
    for assignment, solutions in assignments:
        new_solutions = []
        for input_objects, output_objects in solutions:
            new_output_objects = []
            for obj in output_objects:
                if obj and obj.is_media():
                    attachment = Attachment(id=obj.url, need_upload=assignment.status == toloka.Assignment.SUBMITTED)
                    found_attachments.append(attachment)
                    if id_to_uploaded_attachment:
                        obj = type(obj)(url=s3.get_url(id_to_uploaded_attachment[attachment.id]))
                new_output_objects.append(obj)
            new_solutions.append((input_objects, tuple(new_output_objects)))
        new_assignments.append((assignment, new_solutions))
    return new_assignments, found_attachments


def _get_media_meta(toloka_client: toloka.TolokaClient, attachments: List[Attachment]) -> List[Attachment]:
    if not attachments:
        return []
    return ThreadPool(50).map(
        lambda attachment: Attachment(
            id=attachment.id, meta=toloka_client.get_attachment(attachment.id), need_upload=attachment.need_upload
        ),
        attachments,
    )


def _upload_media(s3: S3, toloka_client: toloka.TolokaClient, attachments: List[Attachment]):
    attachments = [attachment for attachment in attachments if attachment.need_upload]
    if not attachments:
        return

    def upload(attachment: Attachment) -> int:
        with io.BytesIO() as f:
            toloka_client.download_attachment(attachment_id=attachment.id, out=f)
            f.flush()
            size = f.tell()
            f.seek(0)
            s3.upload(attachment, f)
            return size

    # TODO (TOLOKA-17294): OOM danger
    # Threads count should be determined by available RAM and biggest attachments. Unfortunately, Toloka does not
    # provide information about attachment size in its API. For now, let's assume 2Gb available RAM and 50Mb attachment
    # size.
    threads = 40
    logger.debug(f'uploading {len(attachments)} attachments with {threads} threads')
    sizes = ThreadPool(threads).map(upload, attachments)
    logger.debug(f'{sum(sizes)} bytes were uploaded')



import abc
from dataclasses import dataclass
from typing import List, Dict, Union, Any, Optional, Tuple, Callable

import toloka.client.project.template_builder as tb


@dataclass
class ConditionContext:
    @dataclass
    class ArgData:
        final_name: str
        is_output: bool

    name_to_arg_data: Dict[str, ArgData]

    def get_toloka_data(self, name: str) -> Union[tb.InputData, tb.OutputData]:
        assert name in self.name_to_arg_data, (
            f'no argument "{name}" in condition context, may be you are trying to reference'
            f'argument which is declared later in argument list'
        )
        arg_data = self.name_to_arg_data[name]
        data = tb.OutputData if arg_data.is_output else tb.InputData
        return data(path=arg_data.final_name)


class Condition:
    @abc.abstractmethod
    def to_toloka_condition(self, ctx: ConditionContext) -> tb.conditions.BaseConditionV1:
        ...


@dataclass
class ConditionEquals(Condition):
    what: str
    to: Any

    def to_toloka_condition(self, ctx: ConditionContext) -> tb.conditions.BaseConditionV1:
        return tb.EqualsConditionV1(
            data=ctx.get_toloka_data(self.what),
            to=self.to,
        )


class Consequence:
    ...


@dataclass
class If:
    condition: Condition
    then: Union['If', Consequence]
    else_: Optional[Union['If', Consequence]] = None

    @staticmethod
    def get_consequence(
        branch: Optional[Union['If', Consequence]],
        view_gen: Callable[[Consequence], tb.view.BaseViewV1],
        condition_context: ConditionContext,
    ) -> Optional[Union['If', Consequence]]:
        if branch is None:
            return None
        if isinstance(branch, If):
            return branch.to_toloka_if(view_gen, condition_context)
        if isinstance(branch, Consequence):
            return view_gen(branch)
        assert False, f'unexpected consequence: {branch}'

    @staticmethod
    def generate_list_sequence(name: str, value_consequences: List[Tuple[Any, Consequence]]) -> 'If':
        first_if = None
        current_if = None

        for value, consequence in value_consequences:
            if_ = If(
                condition=ConditionEquals(what=name, to=value),
                then=consequence,
                else_=None,
            )

            if first_if is None:
                first_if = if_

            if current_if is not None:
                current_if.else_ = if_

            current_if = if_

        return first_if

    def to_toloka_if(
        self,
        view_gen: Callable[[Consequence], tb.view.BaseViewV1],
        condition_context: ConditionContext,
    ) -> tb.IfHelperV1:
        return tb.IfHelperV1(
            condition=self.condition.to_toloka_condition(condition_context),
            then=self.get_consequence(self.then, view_gen, condition_context),
            else_=self.get_consequence(self.else_, view_gen, condition_context),
        )
